# coding: utf-8

"""This file is the entry point to the module Crappy.

It imports all the modules and other resources, and defines aliases.
"""

# Importing the modules of crappy
from . import actuator
from . import blocks
from . import camera
from . import inout
from . import lamcube
from . import links
from . import modifier
from . import tool

# Importing other features
from .__version__ import __version__
from ._global import OptionalModule, docs

# Useful aliases
link = links.link
Block = blocks.Block
Actuator = actuator.Actuator
Camera = camera.Camera
InOut = inout.InOut
Modifier = modifier.Modifier
Path = blocks.generator_path.meta_path.Path

# Useful commands
stop = Block.stop_all
prepare = Block.prepare_all
launch = Block.launch_all
start = Block.start_all
renice = Block.renice_all
reset = Block.reset



# coding: utf-8

from typing import Optional
import logging
from multiprocessing import current_process
import numpy as np
from itertools import combinations

from .box import Box
from .spots_boxes import SpotsBoxes
from ...._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")
try:
  from skimage.filters import threshold_otsu
  from skimage.morphology import label
  from skimage.measure import regionprops
except (ModuleNotFoundError, ImportError):
  label = OptionalModule("skimage", "Please install scikit-image to use"
                         "Video-extenso")
  threshold_otsu = regionprops = label


class SpotsDetector:
  """This class detects round spots on a grey level image.

  It takes an image from a :class:`~crappy.tool.camera_config.CameraConfig`
  window as an input of the :meth:`detect_spots` method, and tries to detect
  the requested number of spots on it. It then stores the position and size of
  the detected spots, to pass them later on to the
  :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` along
  with other variables once the CameraConfig window is closed.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               white_spots: bool = False,
               num_spots: Optional[int] = None,
               min_area: int = 150,
               blur: Optional[int] = 5,
               update_thresh: bool = False,
               safe_mode: bool = False,
               border: int = 5) -> None:
    """Sets the arguments.

    Args:
      white_spots: If :obj:`True`, detects white spots over a black background.
        If :obj:`False`, detects black spots over a white background. Also
        passed to the 
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`.
      num_spots: The number of spots to detect, as an :obj:`int` between `1`
        and `4`. If given, will try to detect exactly that number of spots and
        will fail if not enough spots can be detected. If left to :obj:`None`,
        will detect up to `4` spots, but potentially fewer.
      min_area: The minimum area an object should have to be potentially
        detected as a spot. The value is given in pixels, as a surface unit.
        It must of course be adapted depending on the resolution of the camera
        and the size of the spots to detect.
      blur: An :obj:`int`, odd and greater than `1`, defining the size of the
        kernel to use when applying a median blur filter to the image before
        trying to detect spots. Can also be set to :obj:`None`, in which case
        no median blur filter is applied before detecting the spots. Also
        passed to the 
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`.
      update_thresh: If :obj:`True`, the grey level threshold for detecting
        the spots is re-calculated at each new image. Otherwise, the first
        calculated threshold is kept for the entire test. The spots are less
        likely to be lost with adaptive threshold, but the measurement will be
        more noisy. Adaptive threshold may also yield inconsistent results when
        spots are lost. Passed to the 
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` 
        and not used in this class.
      safe_mode: If :obj:`True`, will stop and raise an exception as soon as
        overlapping spots are detected. Otherwise, will first try to reduce the
        detection window to get rid of overlapping. This argument should be
        used when inconsistency in the results may have critical consequences.
        Passed to the 
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` 
        and not used in this class.
      border: When searching for the new position of a spot, will search in the
        last known bounding box of this spot plus a few additional pixels in
        each direction. This argument sets the number of additional pixels to
        use. It should be greater than the expected "speed" of the spots, in
        pixels / frame. But if set too high, noise or other spots might hinder
        the detection. Passed to the 
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` 
        and not used in this class.
    """

    self.white_spots = white_spots
    self.thresh = 0 if white_spots else 255
    self.min_area = min_area
    self.blur = blur
    self.update_thresh = update_thresh
    self.safe_mode = safe_mode
    self.border = border

    self._logger = logging.getLogger(f"{current_process().name}."
                                     f"{type(self).__name__}")

    if num_spots is not None and num_spots not in range(1, 5):
      raise ValueError("num_spots should be either None, 1, 2, 3 or 4 !")
    self.num_spots = num_spots

    self.spots = SpotsBoxes()

  def detect_spots(self,
                   img: np.ndarray,
                   y_orig: int,
                   x_orig: int) -> None:
    """Transforms the image to improve spot detection, detects up to 4 spots
    and return a :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes`
    object containing all the detected spots.

    Args:
      img: The sub-image on which the spots should be detected.
      y_orig: The y coordinate of the top-left pixel of the sub-image on the
        entire image.
      x_orig: The x coordinate of the top-left pixel of the sub-image on the
        entire image.

    Returns:
      A :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` object
      containing all the detected spots.
    """

    # First, blurring the image if asked to
    if self.blur is not None and self.blur > 1:
      img = cv2.medianBlur(img, self.blur)

    # Determining the best threshold for the image
    self.thresh = threshold_otsu(img)

    # Getting all pixels superior or inferior to threshold
    if self.white_spots:
      black_white = img > self.thresh
    else:
      black_white = img <= self.thresh

    # The original image is not needed anymore
    del img

    # Detecting the spots on the image
    props = regionprops(label(black_white))

    # The black and white image is not needed anymore
    del black_white

    # Removing regions that are too small or not circular
    props = [prop for prop in props if prop.solidity > 0.8
             and prop.area > self.min_area]

    # Detecting overlapping spots and storing the ones that should be removed
    to_remove = list()
    for prop_1, prop_2 in combinations(props, 2):
      if self._overlap_bbox(prop_1, prop_2):
        to_remove.append(prop_2 if prop_1.area > prop_2.area else prop_1)

    # Now removing the overlapping spots
    if to_remove:
      self._logger.log(logging.WARNING, "Overlapping spots found, removing the"
                                        " smaller overlapping ones")
    for prop in to_remove:
      props.remove(prop)

    # Sorting the regions by area
    props = sorted(props, key=lambda prop: prop.area, reverse=True)

    # Keeping only the biggest areas to match the target number of spots
    if self.num_spots is not None:
      props = props[:self.num_spots]
    else:
      props = props[:4]

    # Indicating the user if not enough spots were found
    if not props:
      self._logger.log(logging.WARNING, "No spots found !")
      return
    elif self.num_spots is not None and len(props) != self.num_spots:
      self._logger.log(logging.WARNING, f"Expected {self.num_spots} spots, "
                                        f"found only {len(props)}")
      return

    # Replacing the previously detected spots with the new ones
    self.spots.reset()
    for i, prop in enumerate(props):
      # Extracting the properties of interest
      y, x = prop.centroid
      y_min, x_min, y_max, x_max = prop.bbox
      # Adjusting with the offset given as argument
      x += x_orig
      x_min += x_orig
      x_max += x_orig
      y += y_orig
      y_min += y_orig
      y_max += y_orig

      self.spots[i] = Box(x_start=x_min, x_end=x_max,
                          y_start=y_min, y_end=y_max,
                          x_centroid=x, y_centroid=y)

  @staticmethod
  def _overlap_bbox(prop_1, prop_2) -> bool:
    """Determines whether two bboxes are overlapping or not."""

    y_min_1, x_min_1, y_max_1, x_max_1 = prop_1.bbox
    y_min_2, x_min_2, y_max_2, x_max_2 = prop_2.bbox

    return max((min(x_max_1, x_max_2) - max(x_min_1, x_min_2)), 0) * max(
      (min(y_max_1, y_max_2) - max(y_min_1, y_min_2)), 0) > 0



# coding: utf-8

from typing import Callable, Optional, Tuple, Union, Any, Dict, Iterable
from time import sleep
import numpy as np
from multiprocessing import current_process
import logging

from .meta_camera import MetaCamera
from .camera_setting import CameraSetting, CameraBoolSetting, \
  CameraScaleSetting, CameraChoiceSetting

NbrType = Union[int, float]


class Camera(metaclass=MetaCamera):
  """Base class for every Camera object. Implements methods shared by all the
  Cameras, and ensures their dataclass is :class:`~crappy.camera.MetaCamera`.

  The Camera objects are helper classes used by the
  :class:`~crappy.blocks.Camera` Block to interface with cameras.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self, *_, **__) -> None:
    """Simply sets the :obj:`dict` containing the settings, and the name of the
    reserved settings.

    Here, :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting` can
    be added to the camera. It can be done using one of the
    :meth:`add_bool_setting`, :meth:`add_scale_setting`,
    :meth:`add_choice_setting`, :meth:`add_trigger_setting`, or
    :meth:`add_software_roi` methods. Refer to the documentation of these
    methods for more information.
    
    .. versionchanged:: 2.0.0 now accepts *args* and *kwargs*
    """

    self.settings: Dict[str, CameraSetting] = dict()

    # These names are reserved for special settings
    self.trigger_name = 'trigger'
    self.roi_x_name = 'ROI_x'
    self.roi_y_name = 'ROI_y'
    self.roi_width_name = 'ROI_width'
    self.roi_height_name = 'ROI_height'
    self._soft_roi_set = False
    self._reserved = (self.trigger_name, self.roi_x_name, self.roi_y_name,
                      self.roi_width_name, self.roi_height_name)

    self._logger: Optional[logging.Logger] = None

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the Camera.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
      
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  def open(self, **kwargs) -> None:
    """This method should initialize the connection to the camera, configure
    the camera, and start the image acquisition.

    Here, :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting` can
    be added to the camera. It can be done using one of the
    :meth:`add_bool_setting`, :meth:`add_scale_setting`,
    :meth:`add_choice_setting`, :meth:`add_trigger_setting`, or
    :meth:`add_software_roi` methods. Refer to the documentation of these
    methods for more information. It is preferable to add the settings during
    :meth:`__init__`, but this is not always possible (e.g. if values read
    from the camera are required to instantiate a setting).

    This method takes as arguments all the kwargs that were passed to the
    :class:`~crappy.blocks.Camera` Block but not used by it. These kwargs can
    have two possible usages. They can be used to parametrize the method, e.g.
    a serial number can be given to choose which camera to open. Alternatively,
    they can be used for adjusting camera settings values. It is also fine not
    to provide any argument to this method.

    When providing a value for a setting, it should be done by giving the kwarg
    ``<setting_name>=<setting_value>`` to the Camera Block, with
    ``<setting_name>`` the exact name given to the setting. It can be desirable
    to provide setting values here in case the display of the
    :class:`~crappy.tool.camera_config.CameraConfig` is disabled
    (``config=False`` set on the Camera Block), or to gain time if the correct
    values are already known.

    Important:
      To effectively set the setting values, the method :meth:`set_all` must be
      called e.g. ``self.set_all(**kwargs)``). It is usually called at the very
      end of the method. This is true even if no value to set was given in the
      kwargs. If :meth:`set_all` is not called, the settings won't actually be
      set on the camera.
    
    .. versionadded:: 1.5.10
    """

    self.set_all(**kwargs)

  def get_image(self) -> Optional[Tuple[Union[Dict[str, Any], float],
                                        np.ndarray]]:
    """Acquires an image and returns it along with its metadata or timestamp.

    This method should return two objects, the second being the image as a
    :mod:`numpy` array. If the first object is a :obj:`float`, it is considered
    as the timestamp associated with the image (as returned by
    :obj:`time.time`). If it is a :obj:`dict`, it should contain the metadata
    associated with the image. It is also fine for this method to return
    :obj:`None` if no image could be acquired.

    If metadata is returned, it should contain at least the ``'t(s)'`` and
    ``'ImageUniqueID'`` keys, containing the timestamp as a :obj:`float` (as
    returned by :obj:`time.time`) and the frame number as an :obj:`int`. Any
    other field can be provided. This metadata is used by the
    :class:`~crappy.blocks.camera_processes.ImageSaver` class and saved along
    with the frames if the ``record_images`` argument of the
    :class:`~crappy.blocks.Camera` Block is :obj:`True`. The keys should
    preferably be valid EXIF tags, so that the information can be embedded in
    the recorded images.

    If only a timestamp is provided, a metadata :obj:`dict` is built internally
    but the user has no control over it.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, "The get_img method was called but is not "
                              "defined !\nNo image can be acquired if this "
                              "method isn't defined")
    sleep(1)
    return

  def close(self) -> None:
    """This method should perform any action required for properly stopping the
    image acquisition and/or closing the connection to the camera.

    This step is usually extremely important in order for the camera resources
    to be released. Otherwise, it might be impossible to re-open the camera
    from Crappy without resetting the hardware connection with it.

    .. versionadded:: 1.5.10
    """

    ...

  def add_bool_setting(self,
                       name: str,
                       getter: Optional[Callable[[], bool]] = None,
                       setter: Optional[Callable[[bool], None]] = None,
                       default: bool = True) -> None:
    """Adds a boolean setting, whose value is either :obj:`True` or
    :obj:`False`.

    It creates an instance of
    :class:`~crappy.camera.meta_camera.camera_setting.CameraBoolSetting` using
    the provided arguments.

    If a :class:`~crappy.tool.camera_config.CameraConfig` window is displayed
    (``config=True`` set on the Camera Block), this setting will appear as a
    checkbox.

    Args:
      name: The name of the setting, that will be displayed in the
        configuration window and allows to access the setting directly with
        ``self.<name>``. Also the name to use for setting the value as a kwarg
        of the Camera Block (``<name>=<value>``).
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting.
    
    .. versionadded:: 1.5.10
    """

    # Checking if the given name is valid
    if name in self._reserved:
      raise ValueError(f"The name {name} is reserved for a different type of "
                       f"setting ! !")
    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.log(logging.INFO, f"Adding the {name} bool setting")
    self.settings[name] = CameraBoolSetting(name, getter, setter, default)

  def add_scale_setting(self,
                        name: str,
                        lowest: NbrType,
                        highest: NbrType,
                        getter: Optional[Callable[[], NbrType]] = None,
                        setter: Optional[Callable[[NbrType], None]] = None,
                        default: Optional[NbrType] = None,
                        step: Optional[NbrType] = None) -> None:
    """Adds a scale setting, whose value is an :obj:`int` or a :obj:`float`
    lying between two boundaries.

    It creates an instance of
    :class:`~crappy.camera.meta_camera.camera_setting.CameraScaleSetting` using
    the provided arguments.

    If a :class:`~crappy.tool.camera_config.CameraConfig` window is displayed
    (``config=True`` set on the Camera Block), this setting will appear as a
    slider.

    Note:
      If any of ``lowest`` or ``highest`` is a :obj:`float`, then the setting
      is considered to be of type float and can take float values. Otherwise,
      it is considered of type :obj:`int` and can only take integer values.

    Args:
      name: The name of the setting, that will be displayed in the
        configuration window and allows to access the setting directly with
        ``self.<name>``. Also the name to use for setting the value as a kwarg
        of the Camera Block (``<name>=<value>``).
      lowest: The lowest possible value for the setting.
      highest: The highest possible value for the setting.
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting. If not given, will
        be the average of ``lowest`` and ``highest``.
      step: The step value for the variation of the setting values.

    .. versionadded:: 1.5.10
    .. versionchanged:: 2.0.0 add the *step* argument
    """

    # Checking if the given name is valid
    if name in self._reserved:
      raise ValueError(f"The name {name} is reserved for a different type of "
                       f"setting ! !")
    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.log(logging.INFO, f"Adding the {name} scale setting")
    self.settings[name] = CameraScaleSetting(name, lowest, highest, getter,
                                             setter, default, step)

  def add_choice_setting(self,
                         name: str,
                         choices: Iterable[str],
                         getter: Optional[Callable[[], str]] = None,
                         setter: Optional[Callable[[str], None]] = None,
                         default: Optional[str] = None) -> None:
    """Adds a choice setting, that can take a limited number of predefined
    :obj:`str` values.

    It creates an instance of
    :class:`~crappy.camera.meta_camera.camera_setting.CameraChoiceSetting`
    using the provided arguments.

    If a :class:`~crappy.tool.camera_config.CameraConfig` window is displayed
    (``config=True`` set on the Camera Block), this setting will appear as a
    set of radio buttons.

    Args:
      name: The name of the setting, that will be displayed in the
        configuration window and allows to access the setting directly with
        ``self.<name>``. Also the name to use for setting the value as a kwarg
        of the Camera Block (``<name>=<value>``).
      choices: An iterable (like a :obj:`tuple` or a :obj:`list`) containing
        the possible values for the setting.
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting. If not given, will
        be the fist item in ``choices``.

    .. versionadded:: 1.5.10
    """

    # Checking if the given name is valid
    if name in self._reserved:
      raise ValueError(f"The name {name} is reserved for a different type of "
                       f"setting ! !")
    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.log(logging.INFO, f"Adding the {name} choice setting")
    self.settings[name] = CameraChoiceSetting(name, tuple(choices), getter,
                                              setter, default)

  def add_trigger_setting(self,
                          getter: Optional[Callable[[], str]] = None,
                          setter: Optional[Callable[[str], None]] = None
                          ) -> None:
    """Adds a specific setting for controlling the trigger mode of the camera.
    The reserved name for this setting is ``'trigger'``.

    It creates an instance of
    :class:`~crappy.camera.meta_camera.camera_setting.CameraChoiceSetting`
    using a reserved name, and predefined choices and default.

    This setting is mainly intended for cameras that can run either in free run
    mode or in hardware trig mode. The three possible choices for this setting
    are :
    ::

      'Free run', 'Hdw after config', 'Hardware'

    Default is ``'Free run'``.

    The setter method is expected to set the camera to free run mode in
    ``'Free run'`` and ``'Hdw after config'`` choices, and to hardware trigger
    mode in the ``'Hardware'`` choice. The getter method should return either
    ``'Free run'`` or ``'Hdw after config'`` in free run mode, depending on the
    last set value for the setting, and ``'Hardware'`` in hardware trig mode.
    It can also be left to :obj:`None`.

    The rationale behind the ``'Hdw after config'`` choice is to allow the user
    to tune settings in the :class:`~crappy.tool.camera_config.CameraConfig`
    window with the camera in free run mode, and to switch afterward to the
    hardware trigger mode once the test begins. It proves extremely useful if
    the hardware triggers are generated from Crappy, as they're not started yet
    when the configuration window is running.

    Args:
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
    
    .. versionadded:: 2.0.0
    """

    if self.trigger_name in self.settings:
      raise ValueError("There can only be one trigger setting per camera !")

    self.log(logging.INFO, f"Adding the '{self.trigger_name}' trigger setting")
    self.settings[self.trigger_name] = CameraChoiceSetting(
      name=self.trigger_name, choices=('Free run',
                                       'Hdw after config',
                                       'Hardware'),
      getter=getter, setter=setter, default='Free run')

  def add_software_roi(self, width: int, height: int) -> None:
    """Creates the settings needed for generating a software ROI.

    The ROI is a rectangular area defining which part of the image to keep and
    return to the :class:`~crappy.blocks.Camera` Block. It can be tuned by the
    user by setting the position of the upper left corner as well as the width
    and the height of the rectangular box.

    Using a ROI reduces the size of the image for processing, displaying and
    saving, which improves the overall performance.

    This method creates instances of
    :class:`~crappy.camera.meta_camera.camera_setting.CameraScaleSetting`
    using reserved names, and predefined choices and default. It takes as
    arguments the dimensions of the un-cropped image, so that it can adjust
    the boundaries of the sliders. These boundaries can later be adjusted
    using the :meth:`reload_software_roi` method.

    The reserved names for the instantiated settings are ``'ROI_x'``,
    ``'ROI_y'``, ``'ROI_width'``, and ``'ROI_height'``, respectively for the
    `x` and `y` position of the upper-left corner of the ROi and for the width
    and height of the ROI.

    Important:
      To apply the ROI and crop the image to return, it is necessary to call
      the :meth:`apply_soft_roi` method. Example :
      ::

        ...
        frame = self._cam.read()
        return time.time(), self.apply_soft_roi(img)

    Args:
      width: The width of the acquired images, in pixels.
      height: The height of the acquired images, in pixels.
    
    .. versionadded:: 2.0.0
    """

    # Checking that the software ROI setting does not already exist
    if self._soft_roi_set:
      raise ValueError("There can only be one set of software settings per "
                       "camera !")

    # Instantiating the CameraSetting objects
    self.log(logging.INFO, "Adding the software ROI settings")
    self.settings[self.roi_x_name] = CameraScaleSetting(
        name=self.roi_x_name, lowest=0, highest=width - 2, getter=None,
        setter=None, default=0, step=1)
    self.settings[self.roi_y_name] = CameraScaleSetting(
        name=self.roi_y_name, lowest=0, highest=height - 2, getter=None,
        setter=None, default=0, step=1)
    self.settings[self.roi_width_name] = CameraScaleSetting(
        name=self.roi_width_name, lowest=2, highest=width, getter=None,
        setter=None, default=width, step=1)
    self.settings[self.roi_height_name] = CameraScaleSetting(
        name=self.roi_height_name, lowest=2, highest=height, getter=None,
        setter=None, default=height, step=1)

    self._soft_roi_set = True

  def reload_software_roi(self, width: int, height: int) -> None:
    """Updates the software ROI boundaries when the width and/or the height of
    the acquired images change.

    The :meth:`add_software_roi` method should have been called before calling
    this method.

    Args:
      width: The width of the acquired images, in pixels.
      height: The height of the acquired images, in pixels.
    
    .. versionadded:: 2.0.0
    """

    if self._soft_roi_set:
      self.log(logging.DEBUG, "Reloading the software ROI settings")
      self.settings[self.roi_x_name].reload(lowest=0, highest=width-1,
                                            value=0, default=0)
      self.settings[self.roi_y_name].reload(lowest=0, highest=height - 1,
                                            value=0, default=0)
      self.settings[self.roi_width_name].reload(lowest=1, highest=width,
                                                value=width, default=width)
      self.settings[self.roi_height_name].reload(lowest=1, highest=height,
                                                 value=height, default=height)
      self.settings[self.roi_x_name].value = 0
      self.settings[self.roi_y_name].value = 0
      self.settings[self.roi_width_name].value = width
      self.settings[self.roi_height_name].value = height
    else:
      self.log(logging.WARNING, "Cannot reload the software ROI settings as "
                                "they are not defined !")

  def apply_soft_roi(self, img: np.ndarray) -> Optional[np.ndarray]:
    """Takes an image as an input, and crops according to the selected software
    ROI dimensions.

    Might return :obj:`None` in case there's no pixel left on the cropped
    image. Returns the original image if the software ROI settings were not
    defined using :meth:`add_software_roi`.
    
    .. versionadded:: 2.0.0
    """

    if self._soft_roi_set:
      x = self.settings[self.roi_x_name].value
      y = self.settings[self.roi_y_name].value
      width = self.settings[self.roi_width_name].value
      height = self.settings[self.roi_height_name].value

      # Cropping to the requested size
      img = img[y:y+height, x:x+width]
      if img.size:
        return img
      # If there's no pixel left to display, return None
      else:
        return

    # Simply returning the image if the ROI settings were not defined
    else:
      return img

  def set_all(self, **kwargs) -> None:
    """Sets all the setting values on the camera.

    The provided keys of the kwargs should be the names of valid
    :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`, otherwise
    an error is raised.

    For settings that are given in the kwargs, sets them to the given
    corresponding value. For the other settings, sets them to their default
    value.

    This method should be called during the :meth:`open` method, once the
    communication with the camera is established and the settings are
    instantiated. It is usually called at the very end of the method.
    """

    unexpected = tuple(kwarg for kwarg in kwargs if kwarg not in self.settings)
    if unexpected:
      raise ValueError(f'Unexpected argument(s) {", ".join(unexpected)} for '
                       f'camera {type(self).__name__}.')

    self.log(logging.INFO, "Setting all the setting values")
    for name, setting in self.settings.items():
      setting.value = kwargs[name] if name in kwargs else setting.default

  def __getattr__(self, item: str) -> Any:
    """Method for getting the value of a setting directly by calling
    ``self.<setting name>``.

    It is called in case :meth:`__getattribute__` doesn't work properly, and
    tries to return the corresponding setting value."""

    try:
      return self.settings[item].value
    except (AttributeError, KeyError):
      raise AttributeError(f'No attribute nor setting named {item}')

  def __setattr__(self, key: str, val: Any) -> None:
    """Method for setting the value of a setting directly by calling
    ``self.<setting name> = <value>``."""

    if key != 'settings' and key in self.settings:
      self.settings[key].value = val
    else:
      super().__setattr__(key, val)



# coding: utf-8

deprecated_cameras = {'Bispectral': 'BiSpectral',
                      'Cl_camera': 'BaslerIronmanCameraLink',
                      'Fake_camera': 'FakeCamera',
                      'File_reader': 'FileReader',
                      'Camera_gstreamer': 'CameraGstreamer',
                      'Jai8': 'JaiGO5000CPMCL8Bits',
                      'Jai': 'JaiGO5000CPMCL',
                      'Camera_opencv': 'CameraOpencv',
                      'Picamera': 'RaspberryPiCamera',
                      'Seek_thermal_pro': 'SeekThermalPro',
                      'Xiapi': 'XiAPI'}



# coding: utf-8

from typing import Optional, Callable, Union, Tuple, Iterable
import numpy as np
from pathlib import Path

from .camera_processes import GPUCorrelProcess
from .camera import Camera


class GPUCorrel(Camera):
  """This Block can perform GPU-accelerated image correlation on images
  acquired by a :class:`~crappy.camera.Camera` object, and project the result
  on various fields.

  It is mostly used for computing the displacement and the strain over the
  given patch, but other fields are also available. refer to the ``fields`` and
  ``labels`` arguments for more details.

  This Block takes no input :class:`~crappy.links.Link` in a majority of
  situations, and outputs the results of image correlation. It is a subclass of
  the :class:`~crappy.blocks.Camera` Block, and inherits of all its features.
  That includes the possibility to record and to display images in real-time,
  simultaneously to the image acquisition and processing. Refer to the
  documentation of the Camera Block for more information on these features.

  This Block is very similar to the :class:`DISCorrel` Block, except this
  latter uses DISFlow to perform the image correlation and is not
  GPU-accelerated. The :class:`~crappy.blocks.GPUVE` Block also relies on
  GPU-accelerated image correlation for computing the displacement and strain
  on images, but it tracks multiple patches and uses video-extensometry.

  No Region Of Interest can be specified to this Block, so by default the
  correlation is performed on the entire image. It is however possible to set
  a mask, so that only part of the image is considered when running the
  correlation.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               camera: str,
               fields: Union[str, Iterable[str]],
               img_shape: Tuple[int, int],
               img_dtype: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               verbose: int = 0,
               freq: Optional[float] = 200,
               debug: Optional[bool] = False,
               save_images: bool = False,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               labels: Optional[Union[str, Iterable[str]]] = None,
               discard_limit: float = 3,
               discard_ref: int = 5,
               img_ref: Optional[np.ndarray] = None,
               levels: int = 5,
               resampling_factor: float = 2,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               mask: Optional[np.ndarray] = None,
               mul: float = 3,
               res: bool = False,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      camera: The name of the :class:`~crappy.camera.Camera` object to use for
        acquiring the images. Arguments can be passed to this Camera as
        ``kwargs`` of this Block. This argument is ignored if the
        ``image_generator`` argument is provided.
      fields: The several fields to calculate on the acquired images. They
        should be given as an iterable containing :obj:`str`. Each string
        represents one field to calculate, so the more fields are given the
        heavier the computation is. The possible fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        For each field, one single value is computed, corresponding to the
        norm of the field values over the patch area.
      img_shape: The shape of the images returned by the
        :class:`~crappy.camera.Camera` object as a :obj:`tuple` of :obj:`int`.
        It should correspond to the value returned by :obj:`numpy.shape`.
      img_dtype: The `dtype` of the images returned by the
        :class:`~crappy.camera.Camera` object, as a :obj:`str`. It should
        correspond to a valid data type in :mod:`numpy`, e.g. ``'uint8'``.
      transform: A callable taking an image as an argument, and returning a
        transformed image as an output. Allows applying a post-processing
        operation to the acquired images. This is done right after the
        acquisition, so the original image is permanently lost and only the
        transformed image is displayed and/or saved and/or further processed.
        The transform operation is not parallelized, so it might negatively
        affect the acquisition framerate if it is too heavy.
      display_images: If :obj:`True`, displays the acquired images in a
        dedicated window, using the backend given in ``displayer_backend`` and
        at the frequency specified in ``displayer_framerate``. This option
        should be considered as a debug or basic follow-up feature, it is not
        intended to be very fast nor to display high-quality images. The
        maximum resolution of the displayed images in `640x480`, the images
        might be downscaled to fit in this format.
      displayer_backend: The backend to use for displaying the images. Can be
        either ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` (OpenCV)
        or :mod:`matplotlib`. ``'cv2'`` usually allows achieving a higher
        display frequency. Ignored if ``display_images`` is :obj:`False`. If
        not given and ``display_images`` is :obj:`True`, ``'cv2'`` is tried
        first and ``'mpl'`` second, and the first available one is used.
      displayer_framerate: The maximum update frequency of the image displayer,
        as an :obj:`int`. This value usually lies between 5 and 30Hz, the
        default is 5. The achieved update frequency might be lower than
        requested. Ignored if ``display_images`` is :obj:`False`.
      software_trig_label: The name of a label used as a software trigger for
        the :class:`~crappy.camera.Camera`. If given, images will only be
        acquired when receiving data over this label. The received value does
        not matter. This software trigger is not meant to be very precise, it
        is recommended not to rely on it for a trigger frequency greater than
        10Hz, in which case a hardware trigger should be preferred if available
        on the camera.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is displayed, and at level `3` so much information
        is displayed that is slows the code down. This argument allows to
        adjust the precision of the log messages, while the ``debug`` argument
        is for enabling or disabling logging.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      save_images: If :obj:`True`, the acquired images are saved to the folder
        specified in ``save_folder``, in the format specified in
        ``img_extension``, using the backend specified in ``save_backend``, and
        at the frequency specified in ``save_period``. Each image is saved with
        the name : ``<frame_nr>_<timestamp>.<extension>``, and can thus easily
        be identified. Along with the images, a ``metadata.csv`` file records
        the metadata of all the saved images. This metadata is either the one
        returned by the :meth:`~crappy.camera.Camera.get_image` method of the
        :class:`~crappy.camera.Camera` object, or the default one generated in
        the :meth:`~crappy.blocks.Camera.loop` method of the
        :class:`~crappy.blocks.Camera` Block. Depending on the framerate of the
        camera and the performance of the computer, it is not guaranteed that
        all the acquired images will be recorded.
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc. Depending on the used ``save_backend``, some
        extensions might not be available. It is currently not possible to
        customize the save parameters further than choosing the file extension.
        Ignored if ``save_images`` is :obj:`False`.
      save_folder: Path to the folder where to save the images, either as a
        :obj:`str` or as a :obj:`pathlib.Path`. Can be an absolute or a
        relative path, pointing to a folder. If the folder does not exist, it
        will be created (if the user has permission). If the given folder
        already contains a ``metadata.csv`` file (and thus likely images from
        Crappy), images are saved to another folder with the same name except
        a suffix is appended. Ignored if ``save_images`` is :obj:`False`. If
        not provided and ``save_images`` is :obj:`True`, the images are saved
        to the folder ``Crappy_images``, created next to the running script.
      save_period: Must be given as an :obj:`int`. Only one out of that number
        images at most will be saved. Allows to have a known periodicity in
        case the framerate is too high to record all the images. Or simply to
        reduce the number of recorded images if saving them all is not needed.
        Ignored if ``save_images`` is :obj:`False`.
      save_backend: If ``save_images`` is :obj:`True`, the backend to use for
        recording the images. It should be one of:
        ::

          'sitk', 'cv2', 'pil', 'npy'

        They correspond to the modules :mod:`SimpleITK`, :mod:`cv2` (OpenCV),
        :mod:`PIL` (Pillow Fork), and :mod:`numpy`. Note that the ``'npy'``
        backend saves the images as raw :obj:`numpy.array`, and thus ignores
        the ``img_extension`` argument. Depending on the machine, some backends
        may be faster or slower. For using each backend, the corresponding
        Python must of course be installed. If not provided and ``save_images``
        is :obj:`True`, the backends are tried in the same order as given above
        and the first available one is used. ``'npy'`` is always available.
      image_generator: A callable taking two :obj:`float` as arguments and
        returning an image as a :obj:`numpy.array`. **This argument is intended
        for use in the examples of Crappy, to apply an artificial strain on a
        base image. Most users should ignore it.** When given, the ``camera``
        argument is ignored and the images are acquired from the generator. To
        apply a strain on the image, strain values (in `%`) should be sent to
        the Camera Block over the labels ``'Exx(%)'`` and ``'Eyy(%)'``.
      labels: The labels to use for sending data to downstream Blocks. If not
        given, the default labels are ``'t(s)', 'meta'`` followed by the names
        of the given ``fields``. They carry for each image its timestamp, a
        :obj:`dict` containing its metadata, and then for each field the
        computed value as a :obj:`float`. When setting this argument, make sure
        to give at least 2 labels for the time and the metadata, and one label
        per field. The ``'res'`` label containing the residuals if ``res`` is
        :obj:`True` should not be included here, it will be automatically
        added.
      discard_limit: If ``res`` is :obj:`True`, the result of the
        correlation is not sent to the downstream Blocks if the residuals for
        the current image are greater than ``discard_limit`` times the average
        residual for the last ``discard_ref`` images.
      discard_ref: If ``res`` is :obj:`True`, the result of the
        correlation is not sent to the downstream Blocks if the residuals for
        the current image are greater than ``discard_limit`` times the average
        residual for the last ``discard_ref`` images.
      img_ref: A reference image, as a 2D :obj:`numpy.array` with `dtype`
        `float32`. If given, it will be set early on the correlation class and
        the test can start from the first acquired frame. If not given, the
        first acquired image will be set as the reference, which will slow down
        the beginning of the test.
      levels: Number of levels of the pyramid. More levels may help converging
        on images with large strain, but may fail on images that don't contain
        low spatial frequency. Fewer levels mean that the program runs faster.
      resampling_factor: The factor by which the resolution is divided between
        each stage of the pyramid. A low factor ensures coherence between the
        stages, but is more computationally intensive. A high factor allows
        reaching a finer detail level, but may lead to a coherence loss between
        the stages.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`. If
        not provided, the default :ref:`GPU Kernels` are used.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing.
      mask: The mask used for weighting the region of interest on the image. It
        is generally used to prevent unexpected behavior on the border of the
        image. Also allows to select a sub-region of the image if the
        correlation should not be performed on the entire image, as this Block
        does not accept a `patch` argument.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk to go past the solution and to diverge. If it's too
        low, the convergence will be slower and require more iterations. `3`
        was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast.
      res: If :obj:`True`, calculates the residuals after performing the
        correlation and returns the residuals along with the correlation data.
        The residuals are always returned under the label ``'res'``, and this
        label should not be included in the ``labels`` argument.
      **kwargs: Any additional argument will be passed to the
        :class:`~crappy.camera.Camera` object, and used as a kwarg to its
        :meth:`~crappy.camera.Camera.open` method.
    
    .. versionadded:: 1.5.10 
       *transform*, *display_images*, *displayer_backend*, 
       *displayer_framerate*, *freq*, *save_images*, *image_generator*, 
       *levels*, *resampling_factor*, *kernel_file*, *iterations*, *mask*, 
       *mul* and *res* arguments
    .. versionchanged:: 1.5.10 
       renamed *discard_lim* argument to *discard_limit*
    .. versionchanged:: 1.5.10 
       renamed *imgref* argument to *img_ref*
    .. versionremoved:: 1.5.10
       *fps_label*, *ext*, *input_label*, *config* and *cam_kwargs* arguments
    .. versionadded:: 2.0.0 
       *debug*, *software_trig_label*, *img_extension*, *img_shape* and
       *img_dtype* arguments
    .. versionremoved:: 2.0.0 *img_name* argument
    """

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=display_images,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=software_trig_label,
                     display_freq=bool(verbose),
                     freq=freq,
                     debug=debug,
                     save_images=save_images,
                     img_extension=img_extension,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     img_shape=img_shape,
                     img_dtype=img_dtype,
                     **kwargs)

    # Forcing the fields into a list
    if isinstance(fields, str):
      fields = [fields]
    else:
      fields = list(fields)

    # Forcing the labels into a list
    if labels is None:
      self.labels = ['t(s)', 'meta'] + fields
    elif isinstance(labels, str):
      self.labels = [labels]
    else:
      self.labels = list(labels)

    if res:
      self.labels.append('res')
    self._calc_res = res

    # Checking that the number of fields and the number of labels match
    if 2 + len(fields) + int(res) != len(self.labels):
      raise ValueError("The number of fields is inconsistent with the number "
                       "of labels !\nMake sure that the time label was given")

    # These arguments are for the GPUCorrelProcess
    self._discard_limit = discard_limit
    self._discard_ref = discard_ref
    self._img_ref = img_ref
    self._verbose = verbose
    self._levels = levels
    self._resampling_factor = resampling_factor
    self._kernel_file = kernel_file
    self._iterations = iterations
    self._fields = fields
    self._mask = mask
    self._mul = mul

  def prepare(self) -> None:
    """This method mostly calls the :meth:`~crappy.blocks.Camera.prepare`
    method of the parent class.

    In addition to that it instantiates the
    :class:`~crappy.blocks.camera_processes.GPUCorrelProcess` object that
    performs the GPU-accelerated image correlation.
    
    .. versionchanged:: 1.5.5 now accepting args and kwargs
    .. versionchanged:: 1.5.10 not accepting arguments anymore
    """

    # Instantiating the GPUCorrelProcess
    self.process_proc = GPUCorrelProcess(
        discard_limit=self._discard_limit,
        discard_ref=self._discard_ref,
        calc_res=self._calc_res,
        img_ref=self._img_ref,
        verbose=self._verbose,
        levels=self._levels,
        resampling_factor=self._resampling_factor,
        kernel_file=self._kernel_file,
        iterations=self._iterations,
        fields=self._fields,
        mask=self._mask,
        mul=self._mul)

    super().prepare()

  def _configure(self) -> None:
    """No configuration window is available for this Block, so this method was
    left blank."""

    ...



# coding: utf-8

from typing import Optional, Dict, Any
import logging

from .meta_modifier import Modifier


class Diff(Modifier):
  """This Modifier calculates the time derivative of a given label and adds the
  derivative to the returned data.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               label: str,
               time_label: str = 't(s)',
               out_label: Optional[str] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      label: The label whose time derivative to compute.
      time_label: The label carrying the time information.
      out_label: The label carrying the calculated derivative. If not given,
        defaults to ``'d_<label>'``.

    .. versionchanged:: 1.5.10 renamed *time* argument to *time_label*
    """

    super().__init__()
    self._label = label
    self._time_label = time_label
    self._out_label = out_label if out_label is not None else f'd_{label}'

    self._last_t = None
    self._last_val = None

  def __call__(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the data from the upstream Block, updates the derivative value,
    appends it to the data and returns the data.
    
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    # For the first received data, storing it and returning 0
    if self._last_t is None or self._last_val is None:
      self._last_t = data[self._time_label]
      self._last_val = data[self._label]
      data[self._out_label] = 0
      return data

    # Updating the differentiation value with the latest received values
    t = data[self._time_label]
    val = data[self._label]
    diff = (val - self._last_val) / (t - self._last_t)
    # Updating the stored data
    self._last_t = t
    self._last_val = val

    # Returning the updated data
    data[self._out_label] = diff
    self.log(logging.DEBUG, f"Sending {data}")
    return data



# coding: utf-8

from typing import Optional
from time import time, sleep
from math import sqrt, copysign
from threading import Thread, RLock
import logging

from .meta_actuator import Actuator


class FakeStepperMotor(Actuator):
  """This :class:`~crappy.actuator.Actuator` can emulate the
  behavior of a stepper motor used as a linear actuator.

  It can drive the motor either in speed or in position, unlike the other fake
  Actuator :class:`~crappy.actuator.FakeDCMotor` that can only be driven in
  speed. This class can also return the current speed and position of the
  motor.

  Internally, the behavior of the motor is emulated in a separate
  :obj:`~threading.Thread` and is based on the fundamental equations of the
  constantly accelerated linear movement.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self,
               steps_per_mm: float = 100,
               microsteps: int = 256,
               acceleration: float = 20,
               max_speed: float = 10) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      steps_per_mm: The number of full steps needed for the motor to move by 1
        mm. The higher this value, the more precise the motor is but the lower
        the maximum achievable speed.
      microsteps: The number of microsteps into which a full step is divided.
        The higher this value, the more precise the resolution of the motor,
        but the lower the maximum achievable speed.
      acceleration: The maximum acceleration the motor can achieve, in
        mm/s². In this demo Actuator, this parameter only has an incidence
        on the responsiveness of the motor.
      max_speed: The maximum achievable speed of the motor, in mm/s.
    """

    super().__init__()

    # The variables describing the motor
    self._accel: float = abs(acceleration * steps_per_mm * microsteps)
    self._max_speed: float = abs(max_speed * steps_per_mm * microsteps)
    self._steps_per_mm: float = steps_per_mm
    self._microsteps: int = microsteps

    # The variables for driving the motor
    self._speed: float = 0
    self._pos: float = 0
    self._target_pos: Optional[float] = 0
    self._target_speed: Optional[float] = None
    self._t: float = time()

    # Attributes managing the Thread emulating the motor
    self._stop_thread: bool = False
    self._lock = RLock()
    self._thread = Thread(target=self._thread_target)

  def open(self) -> None:
    """Starts the :obj:`~threading.Thread` emulating the stepper motor."""

    self.log(logging.INFO, "Starting the emulation thread")
    self._thread.start()

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """Sets the target position that the motor should reach.

    It first converts the given position from mm to steps. If a speed value is
    given, also sets the maximum speed to the given value.

    Args:
      position: The target position to reach, in mm.
      speed: Optionally, the maximum speed at which to move for moving to the
        target position, in mm/s. If the current and target position are too
        close, the motor might not reach this position during its movement.
    """

    with self._lock:
      self._target_speed = None
      self._target_pos = position * self._steps_per_mm * self._microsteps
      if speed is not None:
        self._max_speed = speed * self._steps_per_mm * self._microsteps
      self.log(logging.DEBUG, f"Set the target position to {self._target_pos}"
                              f", with speed {self._max_speed}")

  def set_speed(self, speed: float) -> None:
    """Sets the target speed that the motor should reach.

    It first converts the speed to steps/s, and also checks that the command is
    consistent with the given maximum speed.

    Args:
      speed: The target speed of the motor, in mm/s. Can be negative.
    """

    with self._lock:
      self._target_pos = None
      speed = speed * self._steps_per_mm * self._microsteps
      if abs(speed) > self._max_speed:
        speed = copysign(self._max_speed, speed)
        self.log(logging.WARNING,
                 f"Target speed {speed} higher than the maximum allowed speed "
                 f"{self._max_speed}, setting to max speed")
      self._target_speed = speed

  def get_speed(self) -> float:
    """Returns the current speed of the motor, in mm/s."""

    with self._lock:
      return self._speed / self._steps_per_mm / self._microsteps

  def get_position(self) -> float:
    """Returns the current position of the motor, in mm."""

    with self._lock:
      return self._pos / self._steps_per_mm / self._microsteps

  def stop(self) -> None:
    """Instantly sets the speed and the speed command to 0."""

    with self._lock:
      self.log(logging.INFO, "Abruptly stopping the emulated stepper motor")
      self._speed = 0
      self._target_pos = None
      self._target_speed = 0

  def close(self) -> None:
    """Stops the :obj:`~threading.Thread` emulating the stepper motor if it was
    started."""

    if self._thread.is_alive():
      self.log(logging.INFO, "Trying to stop the emulation thread")
      self._stop_thread = True
      self._thread.join(0.1)
      if self._thread.is_alive():
        self.log(logging.ERROR, "The emulation thread did not terminate "
                                "properly !")

  def _thread_target(self) -> None:
    """The target for the Thread emulating the stepper motor behavior.

    It updates the speed and position of the motor so that the motor reaches
    the target speed or position in the fastest possible way. It is mostly a
    decision tree managing the many possible cases.
    """

    # Emulating the behavior of the motor until told to stop
    while not self._stop_thread:

      # To avoid spamming the CPU
      sleep(0.001)

      # Using a Lock to avoid race conditions with the main Thread
      with self._lock:

        # Constant that will be useful later
        c1 = self._max_speed ** 2 / self._accel

        # Updating the last timestamp
        t = time()
        delta_t = t - self._t
        self._t = t

        # If above the max speed, ignore the target and decelerate
        if self._speed > self._max_speed:
          # Case when we reach the max speed during delta_t
          if self._speed - self._accel * delta_t < self._max_speed:
            self._pos += int(self._max_speed * delta_t
                             + 0.5 * (self._speed - self._max_speed) ** 2
                             / self._accel)
            self._speed = self._max_speed
            continue

          # Case when we're still not reaching the max speed
          self._pos += int(-self._accel * delta_t ** 2 / 2
                           + self._speed * delta_t)
          self._speed += -self._accel * delta_t
          continue

        # If below the min speed, ignore the target and accelerate
        if self._speed < -self._max_speed:
          # Case when we reach the min speed during delta_t
          if self._speed + self._accel * delta_t > -self._max_speed:
            self._pos += int(-self._max_speed * delta_t
                             - 0.5 * (self._speed + self._max_speed) ** 2
                             / self._accel)
            self._speed = self._max_speed
            continue

          # Case when we're still not reaching the min speed
          self._pos += int(self._accel * delta_t ** 2 / 2
                           + self._speed * delta_t)
          self._speed += self._accel * delta_t
          continue

        # Driving the motor in position mode
        if self._target_pos is not None:

          # Calculating the difference and a useful constant
          diff = self._target_pos - self._pos
          c2 = self._speed ** 2 / 2 / self._accel

          # Nothing to do if we're already stopped at the target position
          if diff == 0 and self._speed == 0:
            continue

          # Far enough from target to have time to reach the max or min speed
          if abs(diff) >= c1 - c2:
            # Accelerating to the max speed then decelerating to 0
            if diff > 0:

              # We first have to reach the max speed
              if self._speed < self._max_speed:
                # Case when we reach the max speed during delta_t
                if self._speed + self._accel * delta_t > self._max_speed:
                  self._pos += int(self._max_speed * delta_t
                                   - 0.5 * (self._max_speed - self._speed) ** 2
                                   / self._accel)
                  self._speed = self._max_speed
                  continue

                # Case when we're still not reaching the max speed
                self._pos += int(self._accel * delta_t ** 2 / 2
                                 + self._speed * delta_t)
                self._speed += self._accel * delta_t
                continue

              # We need to decelerate at a precise moment
              else:
                # Case when we reach the limit position during delta_t
                if (self._pos + int(self._max_speed * delta_t) >
                    self._target_pos - c1 / 2):
                  self._speed += (
                      -self._accel * (delta_t - diff / self._max_speed +
                                      self._max_speed / 2 / self._accel))
                  self._pos += int(
                      self._max_speed * delta_t - 0.5 * self._accel *
                      (delta_t - diff / self._max_speed + self._max_speed / 2
                       / self._accel) ** 2)
                  continue

                # Case when we're still not reaching the limit position
                self._pos += int(self._max_speed * delta_t)
                self._speed = self._max_speed
                continue

            # Decelerating to the max speed then accelerating to 0
            else:
              # We first have to reach the min speed
              if self._speed > -self._max_speed:
                # Case when we reach the min speed during delta_t
                if self._speed - self._accel * delta_t < -self._max_speed:
                  self._pos += int(-self._max_speed * delta_t
                                   + 0.5 * (self._max_speed + self._speed) ** 2
                                   / self._accel)
                  self._speed = -self._max_speed
                  continue

                # Case when we're still not reaching the min speed
                self._pos += int(-self._accel * delta_t ** 2 / 2
                                 + self._speed * delta_t)
                self._speed += -self._accel * delta_t
                continue

              # We need to accelerate at a precise moment
              else:
                # Case when we reach the limit position during delta_t
                if (self._pos + int(-self._max_speed * delta_t) <
                    self._target_pos + c1 / 2):
                  self._speed += (
                      self._accel * (delta_t + diff / self._max_speed +
                                     self._max_speed / 2 / self._accel))
                  self._pos += int(
                      -self._max_speed * delta_t + 0.5 * self._accel *
                      (delta_t + diff / self._max_speed + self._max_speed / 2
                       / self._accel) ** 2)
                  continue

                # Case when we're still not reaching the limit position
                self._pos += int(-self._max_speed * delta_t)
                self._speed = -self._max_speed
                continue

          # Too close to target to have time to reach max or min speed
          else:
            # The criteria differ if the speed is positive or negative
            if self._speed >= 0:

              # First accelerate then decelerate to reach target
              if diff >= c2:

                # Close enough to the target, able to reach it within delta_t
                if diff / delta_t <= self._speed <= self._accel * delta_t:
                  self._speed = 0
                  self._pos = self._target_pos
                  continue

                v_lim = sqrt(self._speed ** 2 / 2 + self._accel * diff)

                # Case when we reach the limit speed during delta_t
                if self._speed + self._accel * delta_t > v_lim:
                  self._pos += int(
                      v_lim * delta_t - 0.5 * (v_lim - self._speed) ** 2 /
                      self._accel - self._accel / 2 *
                      (delta_t - (v_lim - self._speed) / self._accel) ** 2)
                  self._speed += (2 * (v_lim - self._speed)
                                  - self._accel * delta_t)
                  continue

                # Case when we're still not reaching the limit speed
                self._pos += int(0.5 * self._accel * delta_t ** 2
                                 + self._speed * delta_t)
                self._speed += self._accel * delta_t
                continue

              # First decelerate then accelerate to reach target
              else:

                v_lim = - sqrt(self._speed ** 2 / 2 - self._accel * diff)

                # Case when we reach the limit speed during delta_t
                if self._speed - self._accel * delta_t < v_lim:
                  self._pos += int(
                      v_lim * delta_t + 0.5 * (v_lim - self._speed) ** 2 /
                      self._accel + self._accel / 2 *
                      (delta_t - (self._speed - v_lim) / self._accel) ** 2)
                  self._speed += (2 * (v_lim - self._speed)
                                  + self._accel * delta_t)
                  continue

                # Case when we're still not reaching the limit speed
                self._pos += int(0.5 * -self._accel * delta_t ** 2
                                 + self._speed * delta_t)
                self._speed += -self._accel * delta_t
                continue

            # The criteria differ if the speed is positive or negative
            else:

              # First decelerate then accelerate to reach target
              if -diff >= c2:

                # Close enough to the target, able to reach it within delta_t
                if -self._accel * delta_t <= self._speed <= diff / delta_t:
                  self._speed = 0
                  self._pos = self._target_pos
                  continue

                v_lim = - sqrt(self._speed ** 2 / 2 - self._accel * diff)

                # Case when we reach the limit speed during delta_t
                if self._speed - self._accel * delta_t < v_lim:
                  self._pos += int(
                      v_lim * delta_t + 0.5 * (v_lim - self._speed) ** 2 /
                      self._accel + self._accel / 2 *
                      (delta_t - (self._speed - v_lim) / self._accel) ** 2)
                  self._speed += (2 * (v_lim - self._speed) +
                                  self._accel * delta_t)
                  continue

                # Case when we're still not reaching the limit speed
                self._pos += int(0.5 * -self._accel * delta_t ** 2
                                 + self._speed * delta_t)
                self._speed += -self._accel * delta_t
                continue

              # First accelerate then decelerate to reach target
              else:

                v_lim = sqrt(self._speed ** 2 / 2 + self._accel * diff)

                # Case when we reach the limit speed during delta_t
                if self._speed + self._accel * delta_t > v_lim:
                  self._pos += int(
                      v_lim * delta_t - 0.5 * (v_lim - self._speed) ** 2 /
                      self._accel - self._accel / 2 *
                      (delta_t - (v_lim - self._speed) / self._accel) ** 2)
                  self._speed += (2 * (v_lim - self._speed)
                                  - self._accel * delta_t)
                  continue

                # Case when we're still not reaching the limit speed
                self._pos += int(0.5 * self._accel * delta_t ** 2
                                 + self._speed * delta_t)
                self._speed += self._accel * delta_t
                continue

        # Driving the motor in speed mode
        if self._target_speed is not None:

          diff = self._target_speed - self._speed

          # Just updating the position if we're already at the target speed
          if diff == 0:
            self._pos += int(self._speed * delta_t)
            continue

          # Case when we need to decelerate to reach the target speed
          if diff < 0:
            # Case when we reach the target speed during delta_t
            if self._speed - self._accel * delta_t < self._target_speed:
              self._pos += int((self._target_speed * delta_t
                               + 0.5 * diff ** 2 / self._accel))
              self._speed = self._target_speed
              continue

            # Case when we're still not reaching the target speed
            self._pos += int(-self._accel * delta_t ** 2 / 2
                             + self._speed * delta_t)
            self._speed += -self._accel * delta_t
            continue

          # Case when we need to accelerate to reach the target speed
          else:
            # Case when we reach the target speed during delta_t
            if self._speed + self._accel * delta_t > self._target_speed:
              self._pos += int(self._target_speed * delta_t
                               - 0.5 * diff ** 2 / self._accel)
              self._speed = self._target_speed
              continue

            # Case when we're still not reaching the target speed
            self._pos += int(self._accel * delta_t ** 2 / 2
                             + self._speed * delta_t)
            self._speed += self._accel * delta_t
            continue



# coding: utf-8

from typing import Dict, Type

from .fake_camera import FakeCamera
from .file_reader import FileReader
from .opencv_camera_webcam import Webcam
from .raspberry_pi_camera import RaspberryPiCamera
from .seek_thermal_pro import SeekThermalPro
from .ximea_xiapi import XiAPI

from .cameralink import BaslerIronmanCameraLink
from .cameralink import JaiGO5000CPMCL, JaiGO5000CPMCL8Bits

from .meta_camera import Camera, MetaCamera, camera_setting

from platform import system
from subprocess import run
if system() == 'Linux':
  try:
    run(['v4l2-ctl'], capture_output=True)
  except FileNotFoundError:
    from .gstreamer_camera_basic import CameraGstreamer
    from .opencv_camera_basic import CameraOpencv
  else:
    from .gstreamer_camera_v4l2 import CameraGstreamer
    from .opencv_camera_v4l2 import CameraOpencv
else:
  from .gstreamer_camera_basic import CameraGstreamer
  from .opencv_camera_basic import CameraOpencv

from ._deprecated import deprecated_cameras
camera_dict: Dict[str, Type[Camera]] = MetaCamera.classes



# coding: utf-8

from multiprocessing import Pipe
from time import time
from copy import deepcopy
from typing import Callable, Union, Any, Dict, Optional, List, Iterable
from collections import defaultdict
from select import select
from platform import system
from multiprocessing import current_process
import logging

from .._global import LinkDataError

ModifierType = Callable[[Dict[str, Any]], Dict[str, Any]]


class Link:
  """This class is used for transferring information between two instances of
  :class:`~crappy.blocks.Block`.

  The created Link is unidirectional, from the input Block to the output Block.
  Under the hood, a Link is basically a :obj:`multiprocessing.Pipe` with
  extra features.

  Note:
    It is possible to add one or multiple :class:`~crappy.modifier.Modifier` to
    modify the transferred value. The Modifiers should be callables taking a
    :obj:`dict` as argument and returning a :obj:`dict`. They can be functions,
    or preferably children of :class:`~crappy.modifier.Modifier`.
  
  .. versionadded:: 1.4.0
  """

  _count = 0

  def __init__(self,
               input_block,
               output_block,
               modifiers: Optional[List[ModifierType]] = None,
               name: Optional[str] = None) -> None:
    """Sets the instance attributes.

    Args:
      input_block: The Block sending data through the Link.
      output_block: The Block receiving data through the Link.
      modifiers: A :obj:`list` containing callables. If several objects given,
        they will be called in the given order. Refer to
        :class:`~crappy.modifier.Modifier` for more information.
      name: Name of the Link, to differentiate it from the others when
        debugging. If no specific name is given, the Links are numbered in the
        order in which they are instantiated in the script.
    
    .. versionchanged:: 1.5.9 renamed *condition* argument to *conditions*
    .. versionchanged:: 1.5.9 renamed *modifier* argument to *modifiers*
    .. versionremoved:: 2.0.0 *conditions*, *timeout* and *action* arguments
    """

    if modifiers is None:
      modifiers = list()

    # Checking that all the given modifiers are callable
    if modifiers and not all(callable(mod) for mod in modifiers):
      not_callable = [mod for mod in modifiers if not callable(mod)]
      raise TypeError(f"The following objects passed as modifiers are not "
                      f"callable : {not_callable} !")

    self.name = name if name is not None else f'link{self._get_count()}'
    self._in, self._out = Pipe()
    self._modifiers = modifiers

    # Associating the link to the input and output blocks
    input_block.add_output(self)
    output_block.add_input(self)

    self._last_warn = time()
    self._logger: Optional[logging.Logger] = None
    self._system = system()

  def __new__(cls, *args, **kwargs):
    """When instantiating a new Link, increments the Link counter."""

    cls._count += 1
    return super().__new__(cls)

  @classmethod
  def _get_count(cls) -> int:
    """Returns the current number of instantiates Links, as an :obj:`int`."""

    return cls._count

  def log(self, log_level: int, msg: str) -> None:
    """Method for recording log messages from the Link.

    Args:
      log_level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{self.name}")

    self._logger.log(log_level, msg)

  def poll(self) -> bool:
    """Returns :obj:`True` if there's data available for reading.
    
    .. versionadded:: 2.0.0
    """

    return self._in.poll()

  def send(self, value: Dict[str, Any]) -> None:
    """Sends a value from the upstream Block to the downstream Block.

    Before sending, applies the given Modifiers and makes sure there's room in
    the Pipe for sending the data (Linux only).
    """

    # Applying the modifiers to the value to send
    for mod in self._modifiers:
      value = mod(deepcopy(value))
      # No need to continue if there's no value to send anymore
      if value is None:
        return

    if not isinstance(value, dict):
      self.log(logging.ERROR, f"Trying to send object of type {type(value)} "
                              f"instead of dict !")
      raise LinkDataError

    # Finally, sending the dict to the link
    if self._system == 'Linux':
      # Can only check on Linux if a pipe is full
      if select([], [self._out], [], 0)[1]:
        self._out.send(value)
      # Warning in case the pipe is full
      elif time() - self._last_warn > 1:
          self._last_warn = time()
          self.log(logging.WARNING, f"Cannot send the values, the Link is "
                                    f"full !")
    else:
      self._out.send(value)

  def recv(self) -> Dict[str, Any]:
    """Reads a single value from the Link and returns it.

    The read value is the oldest available in the Link, see :meth:`recv_last`
    for reading the newest available value.

    If no data is available in the Link, returns an empty :obj:`dict`.

    Returns:
      A :obj:`dict` whose keys are the labels being sent, and for each key a
      single value (usually a :obj:`float` or a :obj:`str`).
      
    .. versionremoved:: 2.0.0 *blocking* argument
    """

    if self._in.poll():
      return self._in.recv()
    else:
      return dict()

  def recv_last(self) -> Dict[str, Any]:
    """Reads all the available values in the Link, and returns the newest one.

    If no data is available in the Link, returns an empty :obj:`dict`. All the
    data that is not returned is permanently dropped.

    Returns:
      A :obj:`dict` whose keys are the labels being sent, and for each key a
      single value (usually a :obj:`float` or a :obj:`str`).
    
    .. versionremoved:: 2.0.0 *blocking* argument
    """

    data = dict()

    while self._in.poll():
      data = self._in.recv()

    return data

  def recv_chunk(self) -> Dict[str, List[Any]]:
    """Reads all the available values in the Link, and returns them all.

    Returns:
      A :obj:`dict` whose keys are the labels being sent, and for each key a
      :obj:`list` of the received values. The first item in the list is the
      oldest one available in the Link, the last item is the newest available.
    
    .. versionremoved:: 1.5.9 *length* argument
    .. versionadded:: 1.5.9 *blocking* argument
    .. versionremoved:: 2.0.0 *blocking* argument
    """

    ret = defaultdict(list)

    while self._in.poll():
      data = self._in.recv()
      for label, value in data.items():
        ret[label].append(value)

    return dict(ret)


def link(in_block,
         out_block,
         modifier: Optional[Union[Iterable[ModifierType],
                                  ModifierType]] = None,
         name: Optional[str] = None) -> None:
  """Function linking two Blocks, allowing to send data from one to the other.

  It instantiates a :class:`~crappy.links.Link` between two children of
  :class:`~crappy.blocks.Block`.

  The created Link is unidirectional, from the input Block to the output Block.
  Under the hood, a Link is basically a :obj:`multiprocessing.Pipe` with
  extra features.

  Args:
    in_block: The Block sending data through the Link.
    out_block: The Block receiving data through the Link.
    modifier: Either a callable, or an iterable (like a :obj:`list` or a
      :obj:`tuple`) containing callables. If several given (in an iterable),
      they are called in the given order. They should preferably be children of
      :class:`~crappy.modifier.Modifier`. Refer to  the associated
      documentation for more information.
    name: Name of the Link, to differentiate it from the others when debugging.
      If no specific name is given, the Links are numbered in the order in
      which they are instantiated in the script.
      
  .. versionadded:: 1.4.0
  .. versionchanged:: 1.5.9
     explicitly listing the *condition*, *modifier*, *timeout*, *action* and
     *name* arguments
  .. versionremoved:: 2.0.0 *condition*, *timeout* and *action* arguments
  """

  # Forcing the modifiers into lists
  if modifier is not None:
    try:
      iter(modifier)
      modifier = list(modifier)
    except TypeError:
      modifier = [modifier]

  # Actually creating the Link object
  Link(input_block=in_block,
       output_block=out_block,
       modifiers=modifier,
       name=name)



# encoding: utf-8

"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, Extension
from os import popen, walk
import platform

# Getting the current version of Python
py_ver = '.'.join(platform.python_version().split('.')[:2])

# The list of extensions to install
extensions = []

# Now finding the extensions to install
install_camera_link = False
install_py_fgen = False
if platform.system() == "Linux":
  # Find the latest runtime version of SiliconSoftware install
  try:
    cl_path = f"/opt/SiliconSoftware/" \
              f"{sorted(next(walk('/opt/SiliconSoftware/'))[1])[-1]}/lib64/"
  except StopIteration:
    # If the software is installed but not found just set cl_path manually here
    cl_path = None

  if cl_path is not None:
    cl_module = Extension(
      'camera.cameralink.clModule',
      sources=['src/ext/Cl_lib/CameraLink.cpp',
               'src/ext/Cl_lib/pyCameraLink.cpp',
               'src/ext/Cl_lib/clSerial.cpp'],
      extra_compile_args=["-std=c++11"],
      extra_link_args=["-l", f"python{py_ver}", "-L", cl_path, "-l", "display",
                       "-l", "clsersis", "-l", "fglib5"],
      include_dirs=[f'/usr/local/lib/python{py_ver}/dist-packages/numpy/'
                    f'core/include'])

    p = popen("lsmod | grep menable")
    if p.read() and install_camera_link:
      extensions.append(cl_module)

if platform.system() == "Windows":

  py_fgen_module = Extension(
    'tool.pyFgenModule',
    include_dirs=[f"C:\\python{py_ver}\\site-packages\\numpy\\core\\include",
                  "C:\\Program Files (x86)\\IVI Foundation\\VISA\\WinNT\\"
                  "include",
                  "C:\\Program Files\\IVI Foundation\\IVI\\Include"],
    sources=['src/ext/niFgen/pyFgen.cpp'], libraries=["niFgen"],
    library_dirs=["C:\\Program Files\\IVI Foundation\\IVI\\Lib_x64\\msc"],
    extra_compile_args=["/EHsc", "/WX"])

  if install_py_fgen:
    extensions.append(py_fgen_module)

  cl_path = "C:\\Program Files\\SiliconSoftware\\Runtime5.2.1\\"
  cl_module = Extension(
    'tool.clModule',
    include_dirs=[f"{cl_path}include",
                  f"C:\\python{py_ver.replace('.', '')}\\Lib\\site-packages\\"
                  f"numpy\\core\\include"],
    sources=['src/ext/Cl_lib/CameraLink.cpp',
             'src/ext/Cl_lib/pyCameraLink.cpp', 'src/ext/Cl_lib/clSerial.cpp'],
    libraries=["clsersis", "fglib5"], library_dirs=[f"{cl_path}lib\\visualc"],
    extra_compile_args=["/EHsc", "/WX"])

  p = popen('driverquery /NH | findstr "me4"')
  if p.read() and install_camera_link:
    extensions.append(cl_module)

setup(ext_package='crappy', ext_modules=extensions)



# coding: utf-8

"""This file contains the code for binding the C++ Spectrum library in Python.

It is only used by the :class:`~crappy.inout.SpectrumM2I4711` InOut. The
bindings haven't been tested for a while.
"""

from ctypes import *
from sys import platform
from ..._global import OptionalModule

# ======= Constants definition ======

ERRORTEXTLEN = 200
SPC_CHENABLE = 11000
SPC_CARDMODE = 9500
SPC_REC_STD_SINGLE = 0x00000001
SPC_REC_FIFO_SINGLE = 0x00000010
SPC_TIMEOUT = 295130
SPC_TRIG_ORMASK = 40410
SPC_TRIG_ANDMASK = 40430
SPC_TMASK_SOFTWARE = 0x00000001
SPC_CLOCKMODE = 20200
SPC_CM_INTPLL = 0x00000001
SPC_AMP0 = 30010
SPC_SAMPLERATE = 20000
SPC_CLOCKOUT = 20110
SPCM_DIR_PCTOCARD = 0
SPCM_DIR_CARDTOPC = 1
SPCM_BUF_DATA = 1000
SPCM_BUF_ABA = 2000
SPCM_BUF_TIMESTAMP = 3000
SPC_M2CMD = 100
M2CMD_CARD_RESET = 0x00000001
M2CMD_CARD_START = 0x00000004
M2CMD_CARD_ENABLETRIGGER = 0x00000008
M2CMD_CARD_DISABLETRIGGER = 0x00000020
M2CMD_CARD_STOP = 0x00000040
M2CMD_CARD_FLUSHFIFO = 0x00000080
M2CMD_CARD_INVALIDATEDATA = 0x00000100
M2CMD_ALL_STOP = 0x00440060
M2CMD_CARD_WAITPREFULL = 0x00001000
M2CMD_CARD_WAITTRIGGER = 0x00002000
M2CMD_CARD_WAITREADY = 0x00004000
M2CMD_DATA_STARTDMA = 0x00010000
M2CMD_DATA_WAITDMA = 0x00020000
M2CMD_DATA_STOPDMA = 0x00040000
M2CMD_DATA_POLL = 0x00080000
SPC_M2STATUS = 110
SPC_DATA_AVAIL_USER_LEN = 200
SPC_DATA_AVAIL_USER_POS = 201
SPC_DATA_AVAIL_CARD_LEN = 202

# ====== Library binding ========

double_reg = [SPC_SAMPLERATE]  # Place here the registers holding 64 bits data
new_buffer = create_string_buffer  # To allow to create a buffer without ctypes

if "linux" in platform.lower():
  msg = "libspcm_linux.so"
  try:
    mod = cdll.LoadLibrary("libspcm_linux.so")
    is_installed = True
  except OSError:
    mod = OptionalModule(msg)
    is_installed = False
elif "darwin" in platform.lower():
  msg = "libc.so.6"
  try:
    mod = cdll.LoadLibrary("libc.so.6")
    is_installed = True
  except OSError:
    mod = OptionalModule(msg)
    is_installed = False
else:
  msg = "spcm_win64.dll"
  try:
    mod = windll.LoadLibrary("C:\\Windows\\system32\\spcm_win64.dll")
    is_installed = True
  except OSError:
    mod = OptionalModule(msg)
    is_installed = False


class SpectrumError(Exception):
  pass


if is_installed:
  dwGetErrorInfo_i32 = mod.spcm_dwGetErrorInfo_i32
  dwGetErrorInfo_i32.argtype = [
      c_void_p, POINTER(c_uint32), POINTER(c_int32), c_char_p]
  dwGetErrorInfo_i32.restype = c_uint32


  def check(h, code):
    if code == 0:
      return
    sz_error_text_buffer = create_string_buffer(ERRORTEXTLEN)
    dwGetErrorInfo_i32(h, None, None, sz_error_text_buffer)
    vClose(h)
    raise SpectrumError(sz_error_text_buffer.value)


  hOpen = mod.spcm_hOpen
  hOpen.argtype = [c_char_p]
  hOpen.restype = c_void_p

  vClose = mod.spcm_vClose
  vClose.argtype = [c_char_p]
  vClose.restype = None

  my_i64 = c_int64()  # C ints to read args when calling getparam
  my_i32 = c_int32()

  mod.spcm_dwGetParam_i32.argtype = [c_void_p, c_int32, POINTER(c_int32)]
  mod.spcm_dwGetParam_i32.restype = c_uint32

  mod.spcm_dwGetParam_i64.argtype = [c_void_p, c_int32, POINTER(c_int64)]
  mod.spcm_dwGetParam_i64.restype = c_uint32


  def dw_get_param(h, reg):
    if reg in double_reg:
      check(h, mod.spcm_dwGetParam_i64(h, reg, byref(my_i64)))
      return my_i64.value
    else:
      check(h, mod.spcm_dwGetParam_i32(h, reg, byref(my_i32)))
      return my_i32.value


  mod.spcm_dwSetParam_i32.argtype = [c_void_p, c_int32, c_int32]
  mod.spcm_dwSetParam_i32.restype = c_uint32

  mod.spcm_dwSetParam_i64.argtype = [c_void_p, c_int32, c_int64]
  mod.spcm_dwSetParam_i64.restype = c_uint32


  def dw_set_param(h, reg, val):
    if reg in double_reg:
      check(h, mod.spcm_dwSetParam_i64(h, reg, val))
    else:
      check(h, mod.spcm_dwSetParam_i32(h, reg, val))


  mod.spcm_dwDefTransfer_i64.argtype = [c_void_p, c_uint32, c_uint32, c_uint32,
                                        c_void_p, c_uint64, c_uint64]
  mod.spcm_dwDefTransfer_i64.restype = c_uint32


  def dw_def_transfer(h, buff_type, direction, notify_size, buff, offset,
                      buff_size):
    check(h, mod.spcm_dwDefTransfer_i64(h, buff_type, direction, notify_size,
                                        buff, c_uint64(offset),
                                        c_uint64(buff_size)))

else:
  dwGetErrorInfo_i32 = OptionalModule(msg)

  hOpen = OptionalModule(msg)

  vClose = OptionalModule(msg)

  my_i64 = OptionalModule(msg)
  my_i32 = OptionalModule(msg)

  dw_get_param = OptionalModule(msg)

  dw_set_param = OptionalModule(msg)

  dw_def_transfer = OptionalModule(msg)



# coding: utf-8

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Tuple
import logging
import numpy as np

from .overlay_object import Overlay


@dataclass
class Box(Overlay):
  """This class represents a box to be drawn on top of the images of a
  :class:`~crappy.tool.camera_config.CameraConfig` window or
  :class:`~crappy.blocks.camera_processes.Displayer` Process of a
  :class:`~crappy.blocks.Camera` Block.

  It is a child of :class:`~crappy.tool.camera_config.config_tools.Overlay`.

  It can represent either the box drawn when selecting a region, or the
  bounding box of a tracked area.
  
  .. versionadded:: 2.0.0
  """

  x_start: Optional[int] = None
  x_end: Optional[int] = None
  y_start: Optional[int] = None
  y_end: Optional[int] = None

  x_disp: Optional[float] = None
  y_disp: Optional[float] = None

  x_centroid: Optional[float] = None
  y_centroid: Optional[float] = None

  def __post_init__(self) -> None:
    """Needed to have the :obj:`~logging.Logger` of the parent class properly
    initialized."""

    super().__init__()

  def __str__(self) -> str:
    """The string representation of this class, only for debugging."""

    return (f"Box with coordinates ({self.x_start}, {self.y_start}), "
            f"({self.x_end}, {self.y_end})")

  def draw(self, img: np.ndarray) -> None:
    """Draws the Box on top of the given image, and returns the modified image.

    The thickness of the drawn lines adapts to the size of the image, so that
    the lines are always visible even when casting the image to a smaller
    format.
    """

    # First, checking if all points are defined
    if self.no_points():
      self.log(logging.DEBUG, f"Cannot draw {self}, not all points are "
                              f"defined !")

    # Getting the thickness of the lines to draw
    x_top, x_bottom, y_left, y_right = self.sorted()
    max_fact = max(img.shape[0] // 480, img.shape[1] // 640, 1)

    # Drawing the lines on top of the image
    try:
      for line in (line for i in range(max_fact + 1) for line in
                   ((self.y_start + i, slice(x_top, x_bottom)),
                    (self.y_end - i, slice(x_top, x_bottom)),
                    (slice(y_left, y_right), x_top + i),
                    (slice(y_left, y_right), x_bottom - i))):
        img[line] = 255 * int(np.mean(img[line]) < 128)
      self.log(logging.DEBUG, f"Drew {self} on top of the image to display")

    # If anything goes wrong, aborting
    except (Exception,) as exc:
      self._logger.exception("Encountered exception while drawing boxes, "
                             "ignoring", exc_info=exc)

  def update(self, box: Box) -> None:
    """Changes the coordinates of the box to those of another box."""

    self.log(logging.DEBUG, f"Updating {self} to {box}")

    self.x_start = box.x_start
    self.y_start = box.y_start
    self.x_end = box.x_end
    self.y_end = box.y_end

    self.x_disp = box.x_disp
    self.y_disp = box.y_disp

    self.x_centroid = box.x_centroid
    self.y_centroid = box.y_centroid

  def no_points(self) -> bool:
    """Returns whether all four sides of the box are defined or not."""

    return any(point is None for point in (self.x_start, self.x_end,
                                           self.y_start, self.y_end))

  def reset(self) -> None:
    """Resets the sides to :obj:`None`."""

    self.log(logging.DEBUG, f"Resetting {self}")

    self.x_start = None
    self.x_end = None
    self.y_start = None
    self.y_end = None

    self.x_centroid = None
    self.y_centroid = None

  def sorted(self) -> Tuple[int, int, int, int]:
    """Returns the four coordinates but sorted in the order : min x, max x,
    min y, max y."""

    if self.no_points():
      self.log(logging.WARNING, f"Trying to sort the Box, but some of its "
                                f"coordinates are undefined !")
      raise ValueError("Cannot sort, some values are None !")

    x_top = min(self.x_start, self.x_end)
    x_bottom = max(self.x_start, self.x_end)
    y_left = min(self.y_start, self.y_end)
    y_right = max(self.y_start, self.y_end)

    self.log(logging.DEBUG, f"Sorted {self}, returning ({x_top}, {x_bottom}, "
                            f"{y_left}, {y_right})")

    return x_top, x_bottom, y_left, y_right



# coding: utf-8

import numpy as np
from time import time
from typing import List, Optional, Iterable
from dataclasses import dataclass
import logging

from .meta_inout import InOut
from .._global import OptionalModule
try:
  import PyDAQmx
except (ModuleNotFoundError, ImportError):
  PyDAQmx = OptionalModule("PyDAQmx")


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a DAQmx
  channel can have."""

  name: str
  range_num: float = 5
  gain: float = 1
  offset: float = 0
  make_zero: bool = False


class DAQmx(InOut):
  """This class can drive data acquisition hardware from National Instruments.

  It is similar to :class:`~crappy.inout.NIDAQmx` InOut, except it relies on
  the :mod:`PyDAQmx` module. It was written and tested on a USB 6008 DAQ board,
  but should work with other instruments as well.

  Note:
    This class requires the NIDAQmx C driver to be installed, as well as the
    :mod:`PyDAQmx` module.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Daqmx to DAQmx
  """

  def __init__(self,
               device: str = 'Dev1',
               channels: Optional[Iterable[str]] = None,
               gain: Optional[Iterable[float]] = None,
               offset: Optional[Iterable[float]] = None,
               ranges: Optional[Iterable[float]] = None,
               make_zero: Optional[Iterable[bool]] = None,
               sample_rate: float = 10000,
               out_channels: Optional[Iterable[str]] = None,
               out_gain: Optional[Iterable[float]] = None,
               out_offset: Optional[Iterable[float]] = None,
               out_ranges: Optional[Iterable[float]] = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      device: The name of the device to open, as a :obj:`str`.
      channels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        the names of the channels to use as inputs, given as :obj:`str`.
        Typical names for inputs are ``'aiX'```, with `X` an integer.
      gain: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each input channel the gain to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no gain is applied to the measured values.
      offset: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each input channel the offset to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no offset is applied to the measured values.
      ranges: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each input channel the range to set for that channel, as a
        :obj:`float`. The possible range values are :
        ::

          0.5, 1., 2.5, 5.

        If not given, all input channels will be set to the range `5`.
      make_zero: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each input channel a :obj:`bool` indicating whether the channel
        should be zeroed or not. If so, data will be acquired on this channel
        before the test starts, and a compensation value will be deduced so
        that the offset of this channel is `0`. **It will only take effect if
        the** ``make_zero_delay`` **argument of the**
        :class:`~crappy.blocks.IOBlock` **controlling the DAQ is set** ! If not
        given, the channels are by default not zeroed.
      sample_rate: The frequency of the acquisition, as a :obj:`float`. The
        higher this number, the more noise there is on the signal but the
        higher the acquisition frequency.
      out_channels: An iterable (like a :obj:`list` or a :obj:`tuple`)
        containing the names of the channels to use as outputs, given as
        :obj:`str`. Typical names for outputs are ``'aoX'``, with `X` an
        integer.
      out_gain: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each output channel the gain to apply to the command voltage, as a
        :obj:`float`. The set voltage is calculated as follows :
        :math:`set\_voltage = out\_gain * command\_voltage + out\_offset`. If
        not given, no gain is applied to the command values.
      out_offset: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each output channel the offset to apply to the command voltage, as
        a :obj:`float`. The set voltage is calculated as follows :
        :math:`set\_voltage = out\_gain * command\_voltage + out\_offset`. If
        not given, no offset is applied to the command values.
      out_ranges: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each output channel the range to set for that channel, as a
        :obj:`float`. The possible range values are :
        ::

          0.5, 1., 2.5, 5.

        If not given, all output channels will be set to the range `5`.

    Note:
      All the iterables given as arguments for the input channels should have
      the same length, and same for the output channels. If that's not the
      case, all the given iterables are treated as if they had the same length
      as the shortest given one.

    .. versionchanged:: 1.5.10 renamed *range* argument to *ranges*
    .. versionchanged:: 1.5.10 renamed *out_range* argument to *out_ranges*
    .. versionremoved:: 1.5.10 *nperscan* argument
    """

    self._handle = None
    self._out_handle = None

    super().__init__()

    self._device = device
    self._sample_rate = sample_rate

    # Setting the defaults for arguments that are not given
    if channels is None:
      channels = list()
    if out_channels is None:
      out_channels = list()

    if ranges is None:
      ranges = [5 for _ in channels]
    if gain is None:
      gain = [1 for _ in channels]
    if offset is None:
      offset = [0 for _ in channels]
    if make_zero is None:
      make_zero = [False for _ in channels]

    if out_ranges is None:
      out_ranges = [5 for _ in out_channels]
    if out_gain is None:
      out_gain = [1 for _ in out_channels]
    if out_offset is None:
      out_offset = [0 for _ in out_channels]

      # Creating the channel objects
      self._channels = [_Channel(name=chan, range_num=r_num, gain=g,
                                 offset=off, make_zero=make_z)
                        for chan, r_num, g, off, make_z in
                        zip(channels, ranges, gain, offset, make_zero)]
      self.log(logging.DEBUG, f"Input channels: {self._channels}")

      self._out_channels = [_Channel(name=chan, range_num=r_num, gain=g,
                                     offset=off) for chan, r_num, g, off in
                            zip(out_channels, out_ranges,
                                out_gain, out_offset)]
      self.log(logging.DEBUG, f"Output channels: {self._out_channels}")

    self._handle, self._out_handle = None, None
    self._n_reads = PyDAQmx.int32()

  def open(self) -> None:
    """Opens the device and initializes the input and output channels."""

    self.log(logging.INFO, "Opening the connection to the DAQmx device")
    PyDAQmx.DAQmxResetDevice(self._device)

    if self._channels:
      # Opening the device for reading
      self._handle = PyDAQmx.TaskHandle()
      PyDAQmx.DAQmxCreateTask('', PyDAQmx.byref(self._handle))

      # Setting up the input channels
      self.log(logging.INFO, "Setting up the input channels")
      for chan in self._channels:
        self.log(logging.DEBUG, f"Setting up the input channel "
                                f"{self._device}/{chan.name}")
        PyDAQmx.DAQmxCreateAIVoltageChan(self._handle,
                                         f"{self._device}/{chan.name}", '',
                                         PyDAQmx.DAQmx_Val_Cfg_Default, 0,
                                         chan.range_num,
                                         PyDAQmx.DAQmx_Val_Volts, None)

    if self._out_channels:
      # Opening the device for writing
      self._out_handle = PyDAQmx.TaskHandle()
      PyDAQmx.DAQmxCreateTask('', PyDAQmx.byref(self._out_handle))

      # Setting up the output channels
      self.log(logging.INFO, "Setting up the output channels")
      for chan in self._out_channels:
        self.log(logging.DEBUG, f"Setting up the output channel "
                                f"{self._device}/{chan.name}")
        PyDAQmx.DAQmxCreateAOVoltageChan(self._out_handle, 
                                         f"{self._device}/{chan.name}", '', 0, 
                                         chan.range_num, 
                                         PyDAQmx.DAQmx_Val_Volts, None)
        PyDAQmx.DAQmxStartTask(self._out_handle)
  
  def make_zero(self, delay: float) -> None:
    """Overriding the method of the parent class, because the user can choose
    which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation to
    zero for the channels that shouldn't be zeroed.
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:
        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]
  
  def get_data(self) -> List[float]:
    """Creates and starts an acquisition task, and returns the acquired
    values."""

    # Defining the acquisition task and the buffer, and starting it
    PyDAQmx.DAQmxCfgSampClkTiming(self._handle, '', self._sample_rate,
                                  PyDAQmx.DAQmx_Val_Rising,
                                  PyDAQmx.DAQmx_Val_FiniteSamps,
                                  2)
    PyDAQmx.DAQmxStartTask(self._handle)
    data = np.empty(len(self._channels), dtype=np.float64)

    # Reading the acquired values and stopping the task
    t0 = time()
    PyDAQmx.DAQmxReadAnalogF64(self._handle, 1, 10.0,
                               PyDAQmx.DAQmx_Val_GroupByChannel, data,
                               len(self._channels),
                               PyDAQmx.byref(self._n_reads), None)
    PyDAQmx.DAQmxStopTask(self._handle)

    return [t0] + [data[i] * chan.gain + chan.offset
                   for i, chan in enumerate(self._channels)]

  def set_cmd(self, *cmd: float) -> None:
    """Sets the command value on the output channels.

    There should be as many commands as there are output channels. In case the
    numbers of channels and commands don't match, an exception is raised.
    """

    # Applying the gains and offsets to the commands
    out_gains = [chan.gain for chan in self._out_channels]
    out_offsets = [chan.offset for chan in self._out_channels]
    data = np.array(cmd, dtype=np.float64) * out_gains + out_offsets

    # Setting the commands
    PyDAQmx.DAQmxWriteAnalogF64(self._out_handle, 1, 1, 10.0,
                                PyDAQmx.DAQmx_Val_GroupByChannel,
                                data, None, None)

  def close(self) -> None:
    """Closes the processes of the input and output channels."""

    # Stopping and closing the processes for data acquisition
    if self._handle is not None:
      self.log(logging.INFO, "Stopping the input channels")
      PyDAQmx.DAQmxStopTask(self._handle)
      PyDAQmx.DAQmxClearTask(self._handle)

    # Stopping and closing the processes for writing data
    if self._out_handle is not None:
      self.log(logging.INFO, "Stopping the output channels")
      PyDAQmx.DAQmxStopTask(self._out_handle)
      PyDAQmx.DAQmxClearTask(self._out_handle)



# coding: utf-8

from typing import Optional, Callable

from .camera_setting import CameraSetting


class CameraBoolSetting(CameraSetting):
  """Camera setting that can only be :obj:`True` or :obj:`False`.

  It is a child of
  :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0
     renamed from Camera_bool_setting to CameraBoolSetting
  """

  def __init__(self,
               name: str,
               getter: Optional[Callable[[], bool]] = None,
               setter: Optional[Callable[[bool], None]] = None,
               default: bool = True) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    super().__init__(name, getter, setter, default)



# coding: utf-8

import numpy as np
from typing import Optional, Tuple
import logging
from _tkinter import TclError

from .meta_block import Block
from .._global import OptionalModule

plt = OptionalModule('matplotlib.pyplot', lazy_import=True)


class Grapher(Block):
  """This Block can display data in a 2D graph in a persistent way.
  
  The graph is displayed in an independent window. It iis updated each time new
  data is received from upstream Blocks. The displayed data can come from 
  different Blocks.
  
  The user can choose which labels are plotted on the `x` and `y` axes. It is
  therefore possible to plot a label versus time, or a label versus another 
  label. A single graph is displayed, but multiple curves can be plotted on
  this graph.
  
  The Grapher Block is known for being very CPU-intensive. For displaying only 
  the last values of given labels, the :class:`~crappy.blocks.LinkReader` and 
  :class:`~crappy.blocks.Dashboard` Blocks are simpler solutions that go much
  easier on the CPU. 
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               *labels: Tuple[str, str],
               length: int = 0,
               freq: Optional[float] = 2,
               max_pt: int = 20000,
               window_size: Tuple[int, int] = (8, 8),
               window_pos: Optional[Tuple[int, int]] = None,
               interp: bool = True,
               backend: str = "TkAgg",
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      *labels: Each :obj:`tuple` corresponds to a curve to plot, and should
        contain two values: the first will be the label of the `x` values, the
        second the label of the `y` values. There's no limit to the number of
        curves. Note that all the curves are displayed in a same graph.
      length: If `0` the graph is static and displays all data from the start
        of the assay. Else, only displays the last ``length`` received chunks,
        and drops the previous ones.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      max_pt: The maximum number of points displayed on the graph. When
        reaching this limit, the Block deletes one point out of two to avoid
        using too much memory and CPU.
      window_size: The size of the graph, in inches.
      window_pos: The position of the graph in pixels. The first value is for
        the `x` direction, the second for the `y` direction. The origin is the
        top-left corner. Works with multiple screens.
      interp: If :obj:`True`, the data points are linked together by straight
        lines. Else, only the points are displayed.
      backend: The :mod:`matplotlib` backend to use. Performance may vary
        according to the chosen backend. Also, every backend may not be
        available depending on your machine.
      display_freq: if :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    Example:
      ::

        graph = Grapher(('t(s)', 'F(N)'), ('t(s)', 'def(%)'))

      will plot a dynamic graph with two lines plot (:math:`F=f(t)` and
      :math:`def=f(t)`).
      ::

        graph = Grapher(('def(%)', 'F(N)'), length=0)

      will plot a static graph.
      ::

        graph = Grapher(('t(s)', 'F(N)'), length=30)

      will plot a dynamic graph displaying the last 30 chunks of data.

    .. versionadded:: 1.5.6 *verbose* argument
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    .. versionchanged:: 2.0.0 renamed *maxpt* argument to *max_pt*
    """

    super().__init__()
    self.niceness = 10
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._length = length
    self._max_pt = max_pt
    self._window_size = window_size
    self._window_pos = window_pos
    self._interp = interp
    self._backend = backend

    self._labels = labels

    self._ax = None
    self._canvas = None
    self._figure = None
    self._lines = None
    self._factor = None
    self._counter = None
    self._clear_button = None

  def prepare(self) -> None:
    """Configures the figure for displaying data."""

    # Switch to the required backend
    if self._backend:
      self.log(logging.INFO, f"Setting matplotlib backend to {self._backend}")
      plt.switch_backend(self._backend)

    # Create the figure and the subplot
    self._figure = plt.figure(figsize=self._window_size)
    self._canvas = self._figure.canvas
    self._ax = self._figure.add_subplot(111)
    self._figure.canvas.mpl_connect('key_press_event', self._on_press)
    self._ax.set_title('(Press c to clear the graph)',
                       fontsize='small', loc='right')

    # Add the lines or the dots
    self._lines = []
    for _ in self._labels:
      if self._interp:
        self._lines.append(self._ax.plot([], [])[0])
      else:
        self._lines.append(self._ax.plot([], [], 'o', markersize=3)[0])

    # Keep only 1/factor points on each line
    self._factor = [1 for _ in self._labels]
    # Count to drop exactly 1/factor points, no more and no less
    self._counter = [0 for _ in self._labels]

    # Add the legend
    legend = [y for _, y in self._labels]
    plt.legend(legend)
    plt.xlabel(', '.join(set(x for x, _ in self._labels)))
    plt.ylabel(', '.join(set(y for _, y in self._labels)))

    # Add a grid
    plt.grid()

    # Set the dimensions if required
    if self._window_pos:
      mng = plt.get_current_fig_manager()
      mng.window.wm_geometry("+%s+%s" % self._window_pos)

    # Ready to show the window
    self.log(logging.INFO, "Configured the matplotlib window, displaying it")
    plt.tight_layout()
    plt.show(block=False)
    plt.pause(.001)

  def loop(self) -> None:
    """Receives the upcoming data, puts in the display buffer and updates the
    graph."""

    # Receives the data sent by the upstream blocks
    if self.freq >= 10:
      # Assuming that above 10Hz the data won't saturate the links
      data = self.recv_all_data_raw()
    else:
      # Below 10Hz, making sure to flush the pipes at least every 0.1s
      data = self.recv_all_data_raw(delay=1 / 2 / self.freq,
                                    poll_delay=min(0.1, 1 / 2 / self.freq))

    update = False  # Should the graph be updated ?

    # For each curve, looking for the corresponding labels in the received data
    for i, (lx, ly) in enumerate(self._labels):
      x, y = None, None
      for dic in data:
        if lx in dic and ly in dic:
          # Found the corresponding data, getting the new data according to the
          # current resampling factors
          dx = dic[lx][self._factor[i] - self._counter[i] -
                       1::self._factor[i]]
          dy = dic[ly][self._factor[i] - self._counter[i] -
                       1::self._factor[i]]
          # Recalculating the counter
          self._counter[i] = (self._counter[i] +
                              len(dic[lx])) % self._factor[i]
          # Adding the new points to the arrays
          x = np.hstack((self._lines[i].get_xdata(), dx))
          y = np.hstack((self._lines[i].get_ydata(), dy))
          # the graph will need to be updated
          update = True
          # As we found the data, no need to search any further
          break

      # In case no matching labels were found, aborting for this curve
      if x is None:
        continue

      # Adjusting the number of points to remain below the length limit
      if self._length and len(x) >= self._length:
        x = x[-self._length:]
        y = y[-self._length:]

      # Dividing the number of points by two to remain below the maxpt limit
      elif len(x) > self._max_pt:
        self.log(logging.INFO, f"Too many points on the graph "
                               f"{i} ({len(x)}>{self._max_pt})")
        x, y = x[::2], y[::2]
        self._factor[i] *= 2
        self.log(logging.INFO, f"Resampling factor is now {self._factor[i]}")

      # Finally, updating the data on the graph
      self.log(logging.DEBUG, f"Graph data for labels {lx}, {ly}: {x}, {y}")
      self._lines[i].set_xdata(x)
      self._lines[i].set_ydata(y)

    # Updating the graph if necessary
    if update:
      self.log(logging.DEBUG, "Updating the graph")
      self._ax.relim()
      self._ax.autoscale()
      try:
        self._canvas.draw()
      except TclError:
        pass
      self._canvas.flush_events()

  def finish(self) -> None:
    """Closes all the opened :mod:`matplotlib` windows."""

    self.log(logging.INFO, "Closing all matplotlib windows")
    plt.close("all")

  def _on_press(self, event) -> None:
    """Callback catching the keyboard press events.

    When called, resets the display by emptying the data buffers.
    """

    if event.key == 'c':
      for line in self._lines:
        line.set_xdata([])
        line.set_ydata([])
      self.factor = [1 for _ in self._labels]
      self.counter = [0 for _ in self._labels]

      self.log(logging.INFO, "Cleared the matplotlib window")



# coding: utf-8

from time import time, sleep
from typing import Callable, Union, Dict, Optional
from re import split, IGNORECASE, match
import logging
from multiprocessing import current_process

from .meta_path import MetaPath

ConditionType = Callable[[Dict[str, list]], bool]


class Path(metaclass=MetaPath):
  """Base class for all the Generator Path objects.

  The Path object are used by the :class:`~crappy.blocks.Generator` Block to
  generate signals.
  
  .. versionadded:: 1.4.0
  """

  t0: Optional[float] = None
  last_cmd: Optional[float] = None

  def __init__(self, *_, **__) -> None:
    """Here, the arguments given to the Path should be handled.

    If the Path accepts one or more conditions, (e.g. stop conditions for
    switching to the next
    :class:`~crappy.blocks.generator_path.meta_path.Path`), they can be parsed
    here using the :meth:`parse_condition` method. See the code of
    :class:`~crappy.blocks.generator_path.Constant` for an example.

    The ``self.t0`` attribute stores the time when the last command of the
    previous :class:`~crappy.blocks.generator_path.meta_path.Path` was sent,
    and the ``self.last_cmd`` stores the value of the last command of the
    previous :class:`~crappy.blocks.generator_path.meta_path.Path`.

    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    self._logger: Optional[logging.Logger] = None

  def get_cmd(self, data: Dict[str, list]) -> Optional[float]:
    """This method is called by the :class:`~crappy.blocks.Generator` Block to
    get the next command to send.

    It takes as input a :obj:`dict` containing the data received by the
    Generator Block since the last command was sent. Refer to
    :meth:`~crappy.blocks.Block.recv_all_data` for more information on the
    format of this dict.

    This method should output the next command that will be sent by the
    Generator Block, as a numeric value. This value can be calculated based on
    one or several of :

    * the input data (see the code of 
      :class:`~crappy.blocks.generator_path.Conditional`)
    * the current time and the ``self.t0`` attribute (see the code of
      :class:`~crappy.blocks.generator_path.Ramp`)
    * the last sent value, using the ``self.last_cmd`` attribute 
    * any other criteria

    Alternatively, if the :class:`~crappy.blocks.generator_path.meta_path.Path`
    is done and should hand over to the next one, it must raise a
    :exc:`StopIteration` exception. Again, the choice to raise that exception
    can be motivated by the current time, a condition as generated by
    :meth:`parse_condition`, or any other criteria.

    It is also fine for this method to return :obj:`None` if no value should be
    output by the Generator Block for this loop.
    """

    self.log(logging.WARNING, "The get_cmd was called but is not defined ! "
                              "Please define a get_cmd method for your "
                              "Generator path ! Returning the last sent "
                              "command")
    sleep(1)
    return self.last_cmd

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the Path.

    Also instantiates the :obj:`~logging.Logger` when logging the first
    message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  def parse_condition(self,
                      condition: Optional[Union[str, ConditionType]]
                      ) -> ConditionType:
    """This method returns a function allowing to check whether a given 
    condition is met or not.
    
    This returned function takes as an input a :obj:`dict` containing the data
    received by the :class:`~crappy.blocks.Generator` Block since it sent the
    last command. See :meth:`~crappy.blocks.Block.recv_all_data` for
    information on the structure of this data. Based on the input data, it
    returns :obj:`True` if the condition is met, and :obj:`False` otherwise.
    
    The condition can be given already as a function (or a callable), as 
    :obj:`None` or more conveniently as a :obj:`str` to parse. If it is given 
    as a function / callable, this callable is directly returned. If it is
    given as :obj:`None`, a function always returning :obj:`False` is returned.

    If the condition is given as a string, the supported condition types are :
    ::

      '<var> > <threshold>'
      '<var> < <threshold>'
      'delay = <your_delay>'

    With ``<var>``, ``<threshold>`` and ``<your_delay>`` to be replaced
    respectively with the label on which the condition applies, the threshold
    for the condition to become :obj:`True`, and the delay before switching to
    the next :class:`~crappy.blocks.generator_path.meta_path.Path`.

    In the case when a :obj:`str` to parse is given as the condition, a
    function performing the check is generated and returned. This way, the user
    doesn't have to understand the internals of data transfers in Crappy to
    handle custom conditions.
    """

    if not isinstance(condition, str):
      # First case, the condition is None
      if condition is None:
        self.log(logging.DEBUG, "Condition is None")

        def cond(_: Dict[str, list]) -> bool:
          """Condition always returning False."""

          return False

        return cond
      # Second case, the condition is already a Callable
      elif isinstance(condition, Callable):
        self.log(logging.DEBUG, "Condition is a callable")
        return condition

    # Third case, the condition is a string containing '<'
    if '<' in condition:
      self.log(logging.DEBUG, "Condition is of type var < thresh")
      var, thresh = split(r'\s*<\s*', condition)

      # Return a function that checks if received data is inferior to threshold
      def cond(data: Dict[str, list]) -> bool:
        """Condition checking that the label values are below a given
        threshold."""

        if var in data:
          return any((val < float(thresh) for val in data[var]))
        return False

      return cond

    # Fourth case, the condition is a string containing '>'
    elif '>' in condition:
      self.log(logging.DEBUG, "Condition is of type var > thresh")
      var, thresh = split(r'\s*>\s*', condition)

      # Return a function that checks if received data is superior to threshold
      def cond(data: Dict[str, list]) -> bool:
        """Condition checking that the label values are above a given
        threshold."""

        if var in data:
          return any((val > float(thresh) for val in data[var]))
        return False

      return cond

    # Fifth case, it is a delay condition
    elif match(r'delay', condition, IGNORECASE) is not None:
      self.log(logging.DEBUG, "Condition is of type delay=xx")
      delay = float(split(r'=\s*', condition)[1])

      # Return a function that checks if the delay is expired
      def cond(_: Dict[str, list]) -> bool:
        """Condition checking if a given delay is expired."""

        return time() - self.t0 > delay

      return cond

    # Otherwise, it's an invalid syntax
    else:
      raise ValueError("Wrong syntax for the condition, please refer to the "
                       "documentation")



# coding: utf-8

from typing import Optional, Callable, Union, Tuple, Iterable
import numpy as np
from pathlib import Path

from .camera_processes import VideoExtensoProcess
from .camera import Camera
from ..tool.camera_config import VideoExtensoConfig, SpotsDetector
from .._global import CameraConfigError


class VideoExtenso(Camera):
  """This Block can perform video-extensometry on images acquired by a
  :class:`~crappy.camera.Camera` object, by tracking spots on the images.

  It takes no input :class:`~crappy.links.Link` in a majority of situations,
  and outputs the results of the video-extensometry. It is a subclass of the
  :class:`~crappy.blocks.Camera` Block, and inherits of all its features. That
  includes the possibility to record and to display images in real-time,
  simultaneously to the image acquisition and processing. Refer to the
  documentation of the Camera Block for more information on these features.

  This Block is quite similar to the :class:`~crappy.blocks.DICVE` Block,
  except this latter tracks patches with a texture instead of spots. Both
  Blocks output similar information, although the default labels and the data
  format are slightly different. See the ``labels`` argument for more detail on
  the output values. Similar to the DICVE, the :class:`~crappy.blocks.GPUVE`
  Block also performs video-extensometry based on GPU-accelerated image
  correlation.

  Similar to the :class:`~crappy.tool.camera_config.CameraConfig` window that
  can be displayed by the Camera Block, this Block can display a
  :class:`~crappy.tool.camera_config.VideoExtensoConfig` window before the test
  starts. Here, the user can also detect and select the spots to track. It is
  currently not possible to specify the coordinates of the spots to track as an
  argument, so the use of the configuration window is mandatory. This might
  change in the future.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Video_extenso to VideoExtenso
  """

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               display_freq: bool = False,
               freq: Optional[float] = 200,
               debug: Optional[bool] = False,
               save_images: bool = False,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               img_shape: Optional[Tuple[int, int]] = None,
               img_dtype: Optional[str] = None,
               labels: Optional[Union[str, Iterable[str]]] = None,
               raise_on_lost_spot: bool = True,
               white_spots: bool = False,
               update_thresh: bool = False,
               num_spots: Optional[int] = None,
               safe_mode: bool = False,
               border: int = 5,
               min_area: int = 150,
               blur: Optional[int] = 5,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      camera: The name of the :class:`~crappy.camera.Camera` object to use for
        acquiring the images. Arguments can be passed to this Camera as
        ``kwargs`` of this Block. This argument is ignored if the
        ``image_generator`` argument is provided.
      transform: A callable taking an image as an argument, and returning a
        transformed image as an output. Allows applying a post-processing
        operation to the acquired images. This is done right after the
        acquisition, so the original image is permanently lost and only the
        transformed image is displayed and/or saved and/or further processed.
        The transform operation is not parallelized, so it might negatively
        affect the acquisition framerate if it is too heavy.
      config: If :obj:`True`, a
        :class:`~crappy.tool.camera_config.VideoExtensoConfig` window is
        displayed before the test starts. There, the user can interactively
        adjust the different
        :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`
        available for the selected :class:`~crappy.camera.Camera`, visualize
        the acquired images, and detect and select the spots to track. The test
        starts when closing the configuration window. **It is currently not
        possible to set this argument to** :obj:`False` **!** This might change
        in the future.
      display_images: If :obj:`True`, displays the acquired images in a
        dedicated window, using the backend given in ``displayer_backend`` and
        at the frequency specified in ``displayer_framerate``. This option
        should be considered as a debug or basic follow-up feature, it is not
        intended to be very fast nor to display high-quality images. The
        maximum resolution of the displayed images in `640x480`, the images
        might be downscaled to fit in this format. In addition to the acquired
        frames, the tracked spots are also displayed on the image as an
        overlay.
      displayer_backend: The backend to use for displaying the images. Can be
        either ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` (OpenCV)
        or :mod:`matplotlib`. ``'cv2'`` usually allows achieving a higher
        display frequency. Ignored if ``display_images`` is :obj:`False`. If
        not given and ``display_images`` is :obj:`True`, ``'cv2'`` is tried
        first and ``'mpl'`` second, and the first available one is used.
      displayer_framerate: The maximum update frequency of the image displayer,
        as an :obj:`int`. This value usually lies between 5 and 30Hz, the
        default is 5. The achieved update frequency might be lower than
        requested. Ignored if ``display_images`` is :obj:`False`.
      software_trig_label: The name of a label used as a software trigger for
        the :class:`~crappy.camera.Camera`. If given, images will only be
        acquired when receiving data over this label. The received value does
        not matter. This software trigger is not meant to be very precise, it
        is recommended not to rely on it for a trigger frequency greater than
        10Hz, in which case a hardware trigger should be preferred if available
        on the camera.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      save_images: If :obj:`True`, the acquired images are saved to the folder
        specified in ``save_folder``, in the format specified in
        ``img_extension``, using the backend specified in ``save_backend``, and
        at the frequency specified in ``save_period``. Each image is saved with
        the name : ``<frame_nr>_<timestamp>.<extension>``, and can thus easily
        be identified. Along with the images, a ``metadata.csv`` file records
        the metadata of all the saved images. This metadata is either the one
        returned by the :meth:`~crappy.camera.Camera.get_image` method of the
        :class:`~crappy.camera.Camera` object, or the default one generated in
        the :meth:`~crappy.blocks.Camera.loop` method of the
        :class:`~crappy.blocks.Camera` Block. Depending on the framerate of the
        camera and the performance of the computer, it is not guaranteed that
        all the acquired images will be recorded.
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc. Depending on the used ``save_backend``, some
        extensions might not be available. It is currently not possible to
        customize the save parameters further than choosing the file extension.
        Ignored if ``save_images`` is :obj:`False`.
      save_folder: Path to the folder where to save the images, either as a
        :obj:`str` or as a :obj:`pathlib.Path`. Can be an absolute or a
        relative path, pointing to a folder. If the folder does not exist, it
        will be created (if the user has permission). If the given folder
        already contains a ``metadata.csv`` file (and thus likely images from
        Crappy), images are saved to another folder with the same name except
        a suffix is appended. Ignored if ``save_images`` is :obj:`False`. If
        not provided and ``save_images`` is :obj:`True`, the images are saved
        to the folder ``Crappy_images``, created next to the running script.
      save_period: Must be given as an :obj:`int`. Only one out of that number
        images at most will be saved. Allows to have a known periodicity in
        case the framerate is too high to record all the images. Or simply to
        reduce the number of recorded images if saving them all is not needed.
        Ignored if ``save_images`` is :obj:`False`.
      save_backend: If ``save_images`` is :obj:`True`, the backend to use for
        recording the images. It should be one of:
        ::

          'sitk', 'cv2', 'pil', 'npy'

        They correspond to the modules :mod:`SimpleITK`, :mod:`cv2` (OpenCV),
        :mod:`PIL` (Pillow Fork), and :mod:`numpy`. Note that the ``'npy'``
        backend saves the images as raw :obj:`numpy.array`, and thus ignores
        the ``img_extension`` argument. Depending on the machine, some backends
        may be faster or slower. For using each backend, the corresponding
        Python must of course be installed. If not provided and ``save_images``
        is :obj:`True`, the backends are tried in the same order as given above
        and the first available one is used. ``'npy'`` is always available.
      image_generator: A callable taking two :obj:`float` as arguments and
        returning an image as a :obj:`numpy.array`. **This argument is intended
        for use in the examples of Crappy, to apply an artificial strain on a
        base image. Most users should ignore it.** When given, the ``camera``
        argument is ignored and the images are acquired from the generator. To
        apply a strain on the image, strain values (in `%`) should be sent to
        the Camera Block over the labels ``'Exx(%)'`` and ``'Eyy(%)'``.
      img_shape: The shape of the images returned by the
        :class:`~crappy.camera.Camera` object as a :obj:`tuple` of :obj:`int`.
        It should correspond to the value returned by :obj:`numpy.shape`.
        **This argument is always ignored as** ``config`` **cannot be set to**
        :obj:`False`. This might change in the future.
      img_dtype: The `dtype` of the images returned by the
        :class:`~crappy.camera.Camera` object, as a :obj:`str`. It should
        correspond to a valid data type in :mod:`numpy`, e.g. ``'uint8'``.
        **This argument is always ignored as** ``config`` **cannot be set to**
        :obj:`False`. This might change in the future.
      labels: The labels to use for sending data to downstream Blocks. If not
        given, the default labels are
        ``'t(s)', 'meta', 'Coord(px)', 'Eyy(%)', 'Exx(%)'``. They carry for
        each image its timestamp, a :obj:`dict` containing its metadata, a
        :obj:`list` containing for each spot the coordinates of its center in a
        :obj:`tuple` of :obj:`int`, and the `y` and `x` strain values
        calculated from the displacement and the initial position of the
        spots. If different labels are desired, they should all be provided at
        once in an iterable of :obj:`str` containing the correct number of
        labels (5).
      raise_on_lost_spot: If :obj:`True`, raises an exception when losing the
        spots to track, which stops the test. Otherwise, stops the tracking but
        lets the test go on and silently sleeps.
      white_spots: If :obj:`True`, detects white objects over a black
        background, else black objects over a white background.
      update_thresh: If :obj:`True`, the grey level threshold for detecting the
        spots is re-calculated at each new image. Otherwise, the first
        calculated threshold is kept for the entire test. The spots are less
        likely to be lost with adaptive threshold, but the measurement will be
        more noisy. Adaptive threshold may also yield inconsistent results when
        spots are lost.
      num_spots: The number of spots to detect, as an :obj:`int` between `1`
        and `4`. If given, will try to detect exactly that number of spots and
        will fail if not enough spots can be detected. If left to :obj:`None`,
        will detect up to `4` spots, but potentially fewer.
      safe_mode: If :obj:`True`, the Block will stop and raise an exception as
        soon as overlapping spots are detected. Otherwise, it will first try to
        reduce the detection window to get rid of overlapping. This argument
        should be used when inconsistency in the results may have critical
        consequences.
      border: When searching for the new position of a spot, the Block will
        search in the last known bounding box of this spot plus a few
        additional pixels in each direction. This argument sets the number of
        additional pixels to use. It should be greater than the expected
        "speed" of the spots, in pixels / frame. But if it's set too high,
        noise or other spots might hinder the detection.
      min_area: The minimum area an object should have to be potentially
        detected as a spot. The value is given in pixels, as a surface unit.
        It must of course be adapted depending on the resolution of the camera
        and the size of the spots to detect.
      blur: The size in pixels (as an odd :obj:`int` greater than `1`) of the
        kernel to use when applying a median blur filter to the image before
        the spot detection. If not given, no blurring is performed. A slight
        blur improves the spot detection by smoothening the noise, but also
        takes a bit more time compared to no blurring.
      **kwargs: Any additional argument will be passed to the
        :class:`~crappy.camera.Camera` object, and used as a kwarg to its
        :meth:`~crappy.camera.Camera.open` method.
    
    .. versionadded:: 1.5.10 
       *display_images*, *displayer_backend*, *displayer_framerate*, *freq*,
       *save_images* and *image_generator* arguments
    .. versionremoved:: 1.5.10 
       *ext*, *fps_label*, *wait_l0* and *input_label* arguments
    .. versionchanged:: 1.5.10
       renamed *show_image* argument to *display_images*
    .. versionchanged:: 1.5.10 renamed *end* argument to *raise_on_lost_spot*
    .. versionadded:: 2.0.0
       *debug*, *software_trig_label*, *img_extension*, *img_shape* and
       *img_dtype* arguments
    .. versionremoved:: 2.0.0 *img_name* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__(camera=camera,
                     transform=transform,
                     config=config,
                     display_images=display_images,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=software_trig_label,
                     display_freq=display_freq,
                     freq=freq,
                     debug=debug,
                     save_images=save_images,
                     img_extension=img_extension,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     img_shape=img_shape,
                     img_dtype=img_dtype,
                     **kwargs)

    # Forcing the labels into a list
    if labels is None:
      self.labels = ['t(s)', 'meta', 'Coord(px)', 'Eyy(%)', 'Exx(%)']
    elif isinstance(labels, str):
      self.labels = [labels]
    else:
      self.labels = list(labels)

    # Making sure a coherent number of labels was given
    if len(self.labels) != 5:
      raise ValueError("The number of labels should be 5 !\n"
                       "Make sure that the time label was given")

    self._raise_on_lost_spot = raise_on_lost_spot
    self._spot_detector = SpotsDetector()

    # These arguments are for the SpotsDetector
    self._white_spots = white_spots
    self._num_spots = num_spots
    self._min_area = min_area
    self._blur = blur
    self._update_thresh = update_thresh
    self._safe_mode = safe_mode
    self._border = border

  def prepare(self) -> None:
    """This method mostly calls the :meth:`~crappy.blocks.Camera.prepare`
    method of the parent class.

    In addition to that it instantiates the
    :class:`~crappy.blocks.camera_processes.VideoExtensoProcess` object that
    performs the video-extensometry and the tracking.
    
    .. versionchanged:: 1.5.5 now accepting args and kwargs
    .. versionchanged:: 1.5.10 not accepting arguments anymore
    """

    # Instantiating the SpotsDetector containing the spots to track
    self._spot_detector = SpotsDetector(white_spots=self._white_spots,
                                        num_spots=self._num_spots,
                                        min_area=self._min_area,
                                        blur=self._blur,
                                        update_thresh=self._update_thresh,
                                        safe_mode=self._safe_mode,
                                        border=self._border)

    # Instantiating the VideoExtensoProcess
    self.process_proc = VideoExtensoProcess(
      detector=self._spot_detector,
      raise_on_lost_spot=self._raise_on_lost_spot)

    super().prepare()

  def _configure(self) -> None:
    """This method should instantiate and start the
    :class:`~crappy.tool.camera_config.VideoExtensoConfig` window for
    configuring the :class:`~crappy.camera.Camera` object.

    It should also handle the case when an exception is raised in the
    configuration window.
    """

    config = None

    # Instantiating and starting the configuration window
    try:
      config = VideoExtensoConfig(self._camera, self._log_queue,
                                  self._log_level, self.freq,
                                  self._spot_detector)
      config.main()

    # If an exception is raised in the config window, closing it before raising
    except (Exception,) as exc:
      self._logger.exception("Caught exception in the configuration window !",
                             exc_info=exc)
      if config is not None:
        config.stop()
      raise CameraConfigError

    # Getting the image dtype and shape for setting the shared Array
    if config.shape is not None:
      self._img_shape = config.shape
    if config.dtype is not None:
      self._img_dtype = config.dtype



# coding: utf-8

from typing import Optional
import logging
import logging.handlers
from time import sleep

from .camera_process import CameraProcess
from ...tool.image_processing import VideoExtensoTool, LostSpotError
from ...tool.camera_config import SpotsDetector


class VideoExtensoProcess(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can perform
  video-extensometry by tracking spots on images. It returns the strain and the
  position of the detected spots on the image.

  It is used by the :class:`~crappy.blocks.VideoExtenso` Block to parallelize
  the image processing and the image acquisition. It delegates most of the
  computation to the
  :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`. It is
  from this class that the output values are sent to the downstream Blocks, and
  that the :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` are sent
  to the :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess for
  display.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               detector: SpotsDetector,
               raise_on_lost_spot: bool = True) -> None:
    """Sets the arguments and initializes the parent class.
    
    Args:
      detector: An instance of the
        :class:`~crappy.tool.camera_config.config_tools.SpotsDetector` Tool,
        containing the coordinates of the detected spots to track. This
        argument is passed to the
        :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`
        and not used in this class.
      raise_on_lost_spot: If :obj:`True`, raises an exception when losing the
        spots to track, which stops the test. Otherwise, stops the tracking but
        lets the test go on and silently sleeps.
    """

    super().__init__()

    self._ve: Optional[VideoExtensoTool] = None
    self._detector = detector
    self._raise_on_lost_spot = raise_on_lost_spot
    self._lost_spots = False

  def init(self) -> None:
    """Instantiates the 
    :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` and
    starts tracking the spots."""

    self.log(logging.INFO, "Instantiating the VideoExtenso tool")
    self._ve = VideoExtensoTool(spots=self._detector.spots,
                                thresh=self._detector.thresh,
                                log_level=self._log_level,
                                log_queue=self._log_queue,
                                white_spots=self._detector.white_spots,
                                update_thresh=self._detector.update_thresh,
                                safe_mode=self._detector.safe_mode,
                                border=self._detector.border,
                                blur=self._detector.blur)

    self.log(logging.INFO, "Starting the VideoExtenso spot tracker "
                           "processes")
    self._ve.start_tracking()

  def loop(self) -> None:
    """This method grabs the latest frame and gives it for processing to the
    :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool`. Then
    sends the strain and displacement data to the downstream Blocks.

    If there's no new frame grabbed or if the spots were already lost, doesn't
    do anything. When losing the spots, decides whether to raise an exception
    or not based on the user's choice. Also sends the patches for display to
    the :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess.
    """

    # Processing only if the spots haven't been lost
    if not self._lost_spots:
      
      # Processing the received frame
      try:
        self.log(logging.DEBUG, "Processing the received image")
        data = self._ve.get_data(self.img)
        
        # Sending the results to the downstream Blocks
        if data is not None:
          self.send([self.metadata['t(s)'], self.metadata, *data])

        # Sending the detected spots to the Displayer for display
        self.send_to_draw(self._ve.spots)

      # In case the spots were just lost
      except LostSpotError:
        self.log(logging.INFO, "Spots lost, stopping the spot trackers")
        self._ve.stop_tracking()
        # Raising if specified by the user
        if self._raise_on_lost_spot:
          self.log(logging.ERROR, "Spots lost, stopping the VideoExtenso "
                                  "process")
          raise
        # Otherwise, simply setting a flag so that no additional
        # processing is performed
        else:
          self._lost_spots = True
          self.log(logging.WARNING, "Spots lost, VideoExtenso staying "
                                    "idle until the test ends")
    
    # If the spots were lost, avoid spamming the CPU in vain
    else:
      self.fps_count -= 1
      sleep(0.1)

  def finish(self) -> None:
    """Indicates the 
    :class:`~crappy.tool.image_processing.video_extenso.VideoExtensoTool` to
    stop tracking the spots."""

    if self._ve is not None:
      self.log(logging.INFO, "Stopping the spot trackers before returning")
      self._ve.stop_tracking()



# coding: utf-8

from typing import Union
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  import board
except (ModuleNotFoundError, ImportError):
  board = OptionalModule('board', 'Blinka is necessary to access the GPIO')

try:
  import digitalio
except (ImportError, ModuleNotFoundError):
  digitalio = OptionalModule('digitalio',
                             'Blinka is necessary to access the GPIOs')

gpio_switch_backends = ['Pi4', 'blinka']


class GPIOSwitch(InOut):
  """This class can drive a GPIO high or low on a single board computer.

  When the command value is `1` the GPIO is turned high, when the command is
  `0` it is turned low. Any value other than `0` and `1` raises an error.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Gpio_switch to GPIOSwitch
  """

  def __init__(self,
               pin_out: Union[int, str],
               backend: str) -> None:
    """Checks the validity of the arguments.

    Args:
      pin_out: The GPIO pin to be controlled. On Raspberry Pi, should be an
        :obj:`int` corresponding to a GPIO in BCM convention. With the
        `'blinka'` backend, should be a string holding the name of the pin.
        Refer to blinka's specific documentation for each board for more
        information.
      backend: Should be one of :
        ::

          'Pi4', 'blinka'

        The `'Pi4'` backend only works on the Raspberry Pis. The `'blinka'`
        backend requires installing :mod:`Adafruit-Blinka`, but this module is
        compatible with and maintained on a wide variety of boards.
    
    .. versionadded:: 1.5.10 *backend* and *ft232h_ser_num* arguments
    .. versionremoved:: 2.0.0 *ft232h_ser_num* argument
    """

    self._pin_out = None

    # Checking that the backend is valid
    if not isinstance(backend, str) or backend not in gpio_switch_backends:
      raise ValueError("backend should be in {}".format(gpio_switch_backends))
    self._backend = backend

    # Checking that the given pin is valid for the selected backend
    if backend == 'blinka' and not hasattr(board, pin_out):
      raise TypeError(f'{pin_out} is not a valid pin using the blinka backend '
                      f'on this board !')
    elif backend == 'Pi4' and (not isinstance(pin_out, int) or
                               pin_out not in range(2, 28)):
      raise ValueError('pin_out should be an integer between 2 and 28 when '
                       'using the Pi4 backend !')

    super().__init__()

    # Instantiating the pin object
    if backend == 'Pi4':
      self._pin_out = pin_out
    elif backend == 'blinka':
      self._pin_out = digitalio.DigitalInOut(getattr(board, pin_out))

  def open(self) -> None:
    """Initializes the GPIO."""

    if self._backend == 'Pi4':
      self.log(logging.INFO, "Setting up the GPIOs")
      GPIO.setmode(GPIO.BCM)
      GPIO.setup(self._pin_out, GPIO.OUT)

    elif self._backend == 'blinka':
      self._pin_out.direction = digitalio.Direction.OUTPUT

  def set_cmd(self, *cmd: int) -> None:
    """Drives the GPIO according to the command.

    Args:
      cmd: 1 for driving the GPIO high, 0 for driving it low.
    """

    if cmd[0] not in [0, 1]:
      raise ValueError("The GPIO input can only be 0 or 1")

    if self._backend == 'Pi4':
      GPIO.output(self._pin_out, cmd[0])
    elif self._backend == 'blinka':
      self._pin_out.value = cmd[0]

  def close(self) -> None:
    """Releases the GPIO."""

    if self._backend == 'Pi4':
      self.log(logging.INFO, "Cleaning up the GPIOs")
      GPIO.cleanup()
    elif self._backend == 'blinka' and self._pin_out is not None:
      self._pin_out.deinit()



# coding: utf-8

from time import time
from typing import Optional, List, Iterable
from dataclasses import dataclass
import logging
from  warnings import warn

from .meta_inout import InOut
from ..tool.bindings import comedi_bind as comedi


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Comedi
  channel can have."""

  num: int
  range_num: int = 0
  gain: float = 1
  offset: float = 0
  make_zero: bool = False
  max_data: int = 0
  range_ds: int = 0


class Comedi(InOut):
  """This class can control acquisition boards relying on the Comedi driver.

  It can read data from ADCs on input channels, and set voltages of DACs on
  output channels. Each channel can be tuned independently in terms of range,
  gan, offset, and for input channels it's possible to decide whether they
  should be offset to `0` at the beginning of the test.

  Communicates over serial. This class was implemented and tested on a USB-DUX
  sigma device. IT should work with other devices as well, but that was not
  tested.

  Note:
    This class requires the `libcomedi` library to be installed on the
    computer.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               device: str = '/dev/comedi0',
               sub_device: int = 0,
               channels: Optional[Iterable[int]] = None,
               range_num: Optional[Iterable[int]] = None,
               gain: Optional[Iterable[float]] = None,
               offset: Optional[Iterable[float]] = None,
               make_zero: Optional[Iterable[bool]] = None,
               out_sub_device: int = 1,
               out_channels: Optional[Iterable[int]] = None,
               out_range_num: Optional[Iterable[int]] = None,
               out_gain: Optional[Iterable[float]] = None,
               out_offset: Optional[Iterable[float]] = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      device: The address of the device, as a :obj:`str`.
      sub_device: The id of the subdevice to use for input channels, as an
        :obj:`int`.
      channels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        the indexes of the channels to use as inputs, given as :obj:`int`.
      range_num: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each input channel the index of the range to set for that channel,
        as an :obj:`int`. Refer to the documentation of the board to get the
        correspondence between range indexes and Volts. If not given, all input
        channels will be set to the range `0`.
      gain: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each input channel the gain to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows :
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no gain is applied to the measured values.
      offset: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each input channel the offset to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows :
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no offset is applied to the measured values.
      make_zero: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each input channel a :obj:`bool` indicating whether the channel
        should be zeroed or not. If so, data will be acquired on this channel
        before the test starts, and a compensation value will be deduced so
        that the offset of this channel is `0`. **It will only take effect if
        the** ``make_zero_delay`` **argument of the**
        :class:`~crappy.blocks.IOBlock` **controlling the Comedi is set** ! If
        not given, the channels are by default not zeroed.
      out_sub_device: The id of the subdevice to use for output channels, as an
        :obj:`int`.
      out_channels: An iterable (like a :obj:`list` or a :obj:`tuple`)
        containing the indexes of the channels to use as outputs, given as
        :obj:`int`.
      out_range_num: An iterable (like a :obj:`list` or a :obj:`tuple`)
        containing for each output channel the index of the range to set for
        that channel, as an :obj:`int`. Refer to the documentation of the board
        to get the correspondence between range indexes and Volts. If not
        given, all output channels will be set to the range `0`.
      out_gain: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each output channel the gain to apply to the command voltage, as a
        :obj:`float`. The set voltage is calculated as follows :
        .. math::

          set_voltage = out_gain * command_voltage + out_offset

        If not given, no gain is applied to the command values.
      out_offset: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each output channel the offset to apply to the command voltage, as
        a :obj:`float`. The set voltage is calculated as follows :
        .. math::

          set_voltage = out_gain * command_voltage + out_offset

        If not given, no offset is applied to the command values.

    Note:
      All the iterables given as arguments for the input channels should have
      the same length, and same for the output channels. If that's not the
      case, all the given iterables are treated as if they had the same length
      as the shortest given one.

    .. versionchanged:: 1.5.10
       now explicitly listing the *device*, *subdevice*, *channels*,
       *range_num*, *gain*, *offset*, *make_zero*, *out_subdevice*,
       *out_channels*, *out_range_num*, *out_gain* and *out_offset* arguments
    .. versionchanged:: 2.0.0 renamed *subdevice* argument to *sub_device*
    .. versionchanged:: 2.0.0
       renamed *out_subdevice* argument to *out_sub_device*
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._device = None

    super().__init__()

    self._device_name = device.encode()
    self._sub_device = sub_device
    self._out_sub_device = out_sub_device

    # Setting the defaults for arguments that are not given
    if channels is None:
      channels = list()
    if out_channels is None:
      out_channels = list()

    if range_num is None:
      range_num = [0 for _ in channels]
    if gain is None:
      gain = [1 for _ in channels]
    if offset is None:
      offset = [0 for _ in channels]
    if make_zero is None:
      make_zero = [False for _ in channels]

    if out_range_num is None:
      out_range_num = [0 for _ in out_channels]
    if out_gain is None:
      out_gain = [1 for _ in out_channels]
    if out_offset is None:
      out_offset = [0 for _ in out_channels]

    # Creating the channel objects
    self._channels = [_Channel(num=chan, range_num=r_num, gain=g,
                               offset=off, make_zero=make_z)
                      for chan, r_num, g, off, make_z in
                      zip(channels, range_num, gain, offset, make_zero)]
    self.log(logging.DEBUG, f"Input channels: {self._channels}")

    self._out_channels = [_Channel(num=chan, range_num=r_num, gain=g,
                                   offset=off) for chan, r_num, g, off in
                          zip(out_channels, out_range_num,
                              out_gain, out_offset)]
    self.log(logging.DEBUG, f"Output channels: {self._out_channels}")

    self._device = None

  def open(self) -> None:
    """Opening the Comedi board and setting up the input and output
    channels."""

    # Opening the Comedi board
    self.log(logging.INFO, "Opening the connection to the Comedi device")
    self._device = comedi.comedi_open(self._device_name)

    # Setting up the input channels
    self.log(logging.INFO, "Setting up the input channels")
    for chan in self._channels:
      chan.max_data = comedi.comedi_get_maxdata(self._device, self._sub_device,
                                                chan.num)
      chan.range_ds = comedi.comedi_get_range(self._device, self._sub_device,
                                              chan.num, chan.range_num)

    # Setting up the output channels
    self.log(logging.INFO, "Setting up the output channels")
    for chan in self._out_channels:
      chan.max_data = comedi.comedi_get_maxdata(self._device,
                                                self._out_sub_device,
                                                chan.num)
      chan.range_ds = comedi.comedi_get_range(self._device,
                                              self._out_sub_device,
                                              chan.num, chan.range_num)

  def make_zero(self, delay: float) -> None:
    """Overriding the method of the parent class, because the user can choose
    which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation to
    zero for the channels that shouldn't be zeroed.
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:

        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]

  def set_cmd(self, *cmd: float) -> None:
    """Sets the command value on the output channels.

    There should be as many commands as there are output channels. In case
    there would be fewer commands or channels, the extra commands/channels
    wouldn't be considered/set.
    """

    for val, chan in zip(cmd, self._out_channels):

      # Adjusting with the provided gain and offset
      val = val * chan.gain + chan.offset

      # Converting voltage to numeric
      out_a = comedi.comedi_from_phys(val, chan.range_ds, chan.max_data)

      # Sending the command
      self.log(logging.DEBUG, f"Writing value {out_a} to channel {chan.num}")
      comedi.comedi_data_write(self._device, self._out_sub_device, chan.num,
                               chan.range_num, comedi.AREF_GROUND, out_a)

  def get_data(self) -> List[float]:
    """Reads and returns the value of each channel, adjusted with the given
    gain and offset."""

    data = [time()]

    for chan in self._channels:

      # Reading the numeric values
      data_read = comedi.comedi_data_read(self._device, self._sub_device,
                                          chan.num, chan.range_num,
                                          comedi.AREF_GROUND)
      self.log(logging.DEBUG, f"Read value {data_read} to channel {chan.num}")

      # Converting numeric to a voltage
      val = comedi.comedi_to_phys(data_read, chan.range_ds, chan.max_data)

      # Adjusting with the provided gain and offset
      data.append(val * chan.gain + chan.offset)
    return data

  def close(self) -> None:
    """Closes the Comedi board and warns the user in case of failure."""

    if self._device is not None:
      self.log(logging.INFO, "Closing the connection to the Comedi device")
      if comedi.comedi_close(self._device):
        self.log(logging.WARNING, "Closing device failed !")
